# coding: utf8
from __future__ import absolute_import
import inspect
from farbox_lite.utils import smart_str, smart_unicode
from farbox_lite.utils.functional import cached_property


def get_functions_in_a_file_obj(obj):
    obj_filepath = inspect.getmodule(obj).__file__
    functions = []
    sub_functions = inspect.getmembers(obj, inspect.isfunction)
    for sub_function_name, sub_function in sub_functions:
        if sub_function_name.startswith('_') and not sub_function_name.startswith('__') and not sub_function_name=='_': # 以_开头的函数名，不做处理
            # __开头的比较特殊，为了跟系统内置的名称区分开来，比如__int，可能就是一个 int 的调用
            continue
        _real_function = getattr(sub_function, 'original_func', None)  # 可能是被 wrapper 的函数
        real_function = _real_function or sub_function
        if inspect.getmodule(real_function).__file__==obj_filepath:
            functions.append((sub_function_name, sub_function))

            sub_function.arg_spec = inspect.getargspec(real_function)
            sub_function.doc = (inspect.getdoc(real_function) or '').strip()

            if sub_function_name.startswith('__'):
                func_name = sub_function_name
                if func_name != '_':
                    func_name = sub_function_name.lstrip('_')
                functions.append((func_name, sub_function))
    return dict(functions)




class AttrFunc(object):
    # value.int -> to_int(value)
    # value.int(9) -> to_int(value, 9)
    # value.int() -> to_int(value)
    def __init__(self, func):
        self.func = func
        self.is_attr_func = True

    @cached_property
    def default_value(self):
        try:
            return self.func()
        except:
            return smart_unicode(self.func)


    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return smart_unicode(self.default_value)

    def __unicode__(self):
        return smart_unicode(self.default_value)

    def __str__(self):
        return smart_str(self.default_value)

    def __nonzero__(self):
        return bool(self.func)


    def __add__(self, other):
        # 对相加的特殊处理
        if other is None:
            other = ''
        if isinstance(other, (str, unicode)):
            return '%s%s' % (self.default_value, other)
        else:
            return self.default_value + other

    def __mul__(self, other):
        if other is None:
            return ''
        else:
            return self.default_value * other



# 声明了数据类型，这样可以判断meta_class, 多一个 list 对应
metas = {
    'int': int,
    'float': float,
    'str': str,
    'unicode': unicode,
    'tuple': tuple,
    'list': list,
    'dict': dict,
}

def get_lazy_obj(func, first_arg=None, *default_args, **default_kwargs):
    # 这是对函数做了一个非常特殊的处理，可以被当做变量处理，但是同时，也可以作为函数，继续当做函数调用。
    # 推测函数最后的返回类型
    func_name = func.func_name
    meta = metas.get(func_name.rsplit('-')[-1])
    if not meta:
        func_doc = func.func_doc or ''
        if func_doc: # func_doc的第一行声明了数据类型
            meta = metas.get(func_doc.split('\n')[0].strip())
    if not meta:
        meta = object

    class LazyFunc(meta):
        # value.int -> to_int(value)
        # value.int(9) -> to_int(value, 9)
        # value.int() -> to_int(value)
        def __init__(self):
            self.is_lazy_func_obj = True
            self.func = func
            self.first_arg = first_arg
            meta.__init__(self)

        @cached_property
        def default_value(self):
            if not self.func: # 特殊情况，相当于对常量做了包裹
                return self.first_arg
            if default_args and first_arg is None:
                value = func(*default_args, **default_kwargs)
            else:
                if first_arg is None and not default_args and not default_kwargs:
                    # 比如 [1,2,3].json 的调用，json 本身只接受一个参数（which curried already）
                    #if hasattr(func, 'original_func') and hasattr(func.original_func, 'arg_spec') and func.original_func.arg_spec:
                    #    # 比如 posts.group('??') 这种必须要有参数的
                    #    value = ''
                    #elif hasattr(func, 'arg_spec') and func.arg_spec:
                    #    value = ''
                    try:
                        value = func()
                    except: # 很可能是没有参数是不能运行的状态
                        value = ''
                else:
                    value = func(first_arg, *default_args, **default_kwargs)
            return value


        def __call__(self, *args, **kwargs):
            if not self.func: # 函数为空
                return self.default_value

            if not args and not kwargs:
                return self.default_value
            else: # 有参数传入
                if self.first_arg is None:
                    args = list(args)
                else:
                    args = [self.first_arg] + list(args)
                return self.func(*args, **kwargs)

        def __repr__(self):
            return smart_unicode(self.default_value)

        def __unicode__(self):
            return smart_unicode(self.default_value)

        def __str__(self):
            return smart_str(self.default_value)

        def __nonzero__(self):
            return bool(self.default_value)

        def __eq__(self, other):
            return other == self.default_value

        def __add__(self, other):
            # 对相加的特殊处理
            if other is None:
                other = ''
            if isinstance(other, (str, unicode)):
                return '%s%s' % (self.default_value, other)
            else:
                return self.default_value + other

        def __mul__(self, other):
            if other is None:
                return ''
            else:
                return self.default_value * other

    obj = LazyFunc()

    the_func = func
    if hasattr(func, 'original_func'): # curry 过的
        the_func = func.original_func
    if getattr(the_func, 'is_just_property', None):
        # 仅仅是 property, 比如.length, 这样可以保证数据属性的 ok
        return obj.default_value

    return obj


def value_to_lazy_obj(value):
    return get_lazy_obj(func=None, first_arg=value)






#coding: utf8
from __future__ import absolute_import
from farbox_lite.utils import smart_unicode, get_md5
from farbox_lite.utils.functional import  curry
from flask import _app_ctx_stack, g


local_g = None

class G(object):
    pass


def get_g():
    global local_g
    if not _app_ctx_stack.top:
        # 不是以 app 的形式运行的，可能只是模板的调用
        local_g = local_g or G()
        return local_g
    else:
        return g


def cache_wrapper(func, cache_name):
    # 为了避免冲突，一般cache_name是以'cached_'开头的
    def _func(*args, **kwargs): #the real func
        g = get_g()
        if cache_name == 'no_cache': # 不对结果缓存
            return func(*args, **kwargs)

        # 获得/创建容器
        if not hasattr(g, cache_name):
            setattr(g,  cache_name, {})

        # 准备key值，可以在容器内进行匹配
        values = list(args) + kwargs.keys() + kwargs.values()
        key_list = []
        for value in values:
            if isinstance(value, dict) and '_id' in value:
                key_list.append(smart_unicode(value['_id']))
            else:
                key_list.append(smart_unicode(value))
        key = "-".join(key_list)
        if key:
            key = get_md5(key) # 可以避免key太长了
        else: # 如果是empty，每次的get_hash_key都是不一样的，意味着直接调用cache_result的，都会失效
            key = '_'

        # 如不匹配，则进行计算，并存入容器
        if key not in getattr(g, cache_name): # 仅判断key的存在，不做key对应的值是否存在
            result = func(*args, **kwargs)
            getattr(g, cache_name)[key] = result

        return getattr(g, cache_name).get(key) # 返回匹配值

    # 写入原函数的一些属性
    _func.__name__ = func.__name__
    _func.func_name = func.func_name
    _func.original_func = func

    return _func


def cache_result(*args, **kwargs):
    prefix = 'cached_'

    # @cache_result这样直接用
    if len(args) == 1 and hasattr(args[0], '__call__'):
        func = args[0]
        cache_name = prefix + func.func_name
        return cache_wrapper(func, cache_name)

    # @cache_result('key_value')
    elif len(args) == 1 and isinstance(args[0], basestring):
        cache_name = prefix + args[0]
        return curry(cache_wrapper, cache_name=cache_name)




# coding: utf8
from __future__ import absolute_import
import os
import uuid
from flask import g

from farbox_lite.utils import smart_unicode, get_md5
from farbox_lite.utils.functional import cached_property
from farbox_lite.utils.lazy import get_value_from_data
from farbox_lite.utils.url import join_url
from farbox_lite.utils.html import html_escape, re_html_dom, linebreaks as _linebreaks

from farbox_lite.utils.web.url import get_url_with_prefix
from farbox_lite.utils.web.cache_for_function import cache_result
from farbox_lite.utils.web.request import get_language

from farbox_lite.render.apis.data import get_paginator
from farbox_lite.render.utils.render import render_api_template



class Html(object):
    def __getattr__(self, item):
        # 一些同名的隐射
        if item in ['mobile_headers', 'mobile_header', 'mobile_meta']:
            return self.mobile_metas
        elif item in ['header']:
            return self.headers
        elif item in ['no_inject']:
            return self.set_no_html_inject


    @property
    def a_color(self):
        return self.get_a_color()

    @cached_property
    def prefix_url(self):
        prefix = getattr(g, 'prefix', '')
        if prefix and isinstance(prefix, (str, unicode)):
            prefix = prefix.strip('/')
        return prefix


    @cached_property
    def mobile_metas(self):
        return """
        <meta content="width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0" name="viewport"/>
        <meta content="yes" name="apple-mobile-web-app-capable"/>
        <meta content="black" name="apple-mobile-web-app-status-bar-style"/>
        <meta content="telephone=no" name="format-detection"/>
        <meta name="renderer" content="webkit">\n"""


    def seo(self, keywords=None, description=None):
        if getattr(g, 'seo_header_set', False):
            # 在一个页面内，仅仅能运行一次, 会被 h.headers 调用，如果要使用seo，确保 seo 在之前运行
            return ''
        keywords = keywords or get_value_from_data(g, 'doc.metadata.keywords') or get_value_from_data(g, 'site.configs.keywords')
        if isinstance(keywords, (list, tuple)): # keywords 是一个list
            keywords = [smart_unicode(k) for k in keywords]
            keywords = ', '.join(keywords)
        description = description or get_value_from_data(g, 'doc.metadata.description') or get_value_from_data(g, 'site.configs.description')
        html_content = self.set_metas(keywords=keywords, description=description)
        g.seo_header_set = True
        return html_content


    @cached_property
    def headers(self):
        return self.mobile_metas + self.seo()


    def set_meta(self, key, value):
        # 生成 HTML 的 head_meta
        if not key or not value:
            return ''
        key = smart_unicode(key).replace('"', "'")
        value = html_escape(smart_unicode(value).replace('"', "'"))
        meta_content = '<meta name="%s" content="%s">\n' % (key, value)
        return meta_content

    def set_metas(self, **kwargs):
        html_content = ''
        for k, v in kwargs.items():
            html_content += self.set_meta(k, v)
        return html_content


    @staticmethod
    def set_no_html_inject(status=True):
        # utils.render 的insert_into_header/footer 将会失效
        g.no_html_inject = status
        return ''


    def disable_load(self):
        # 禁用 load 这个函数
        g.disable_load_func = True
        return ''


    @classmethod
    def load(cls, *resource, **kwargs):
        if getattr(g, 'disable_load_func', False): # load函数被禁用了
            return ''

        force_load = kwargs.pop('force', False)

        if not resource:
            return ''


        if len(resource) == 1:
            resource = resource[0]
            if ' ' in resource: # load('a b c')的处理
                resource = resource.split(' ')

        # resource 可以是一个 list，也可以是字符串
        # 预先进行类型判断
        if isinstance(resource, (list, tuple)):
            result = ''
            for child_resource in resource:
                if isinstance(child_resource, (str, unicode)):
                    result += cls.load(child_resource)
            return result
        elif not isinstance(resource, (str, unicode)):
            return ''

        # 确保进入以下流程的都是单个 resource，这样才能达到去重的效果

        #相同的 resource，一个页面内，仅允许载入一次
        loads_in_page = getattr(g, 'loads_in_page', [])
        if not isinstance(loads_in_page, list):
            loads_in_page = []
        if resource in loads_in_page and not force_load:
             #ignore, 页面内已经载入过一次了
            return ''
        else:
            loads_in_page.append(resource)
            g.loads_in_page = loads_in_page

        if not isinstance(resource, (str, unicode)):
            return resource


        resource_path = resource
        if '?' in resource:
            resource_path = resource.split('?')[0]
        ext = os.path.splitext(resource_path)[1].lower()
        if not ext:
            ext = '.%s' % (resource.split('?')[0].split('/')[-1]) # 比如http://fonts.useso.com/css?family=Lato:300

        if ext == '.js' or ext.startswith('.js?') or resource.split('?')[0].endswith('js'):
            content = '<script type="text/javascript" src="%s"></script>' % resource
        elif ext in ['.css', '.less', '.scss', '.sass'] or ext.startswith('.css?') or ext.endswith('?format=css'):
            content = '<link href="%s" type="text/css" rel="stylesheet"/>' % resource
        else:
            content = ''

        return content



    def show_notification(self, message, timeout=0, status=None, position='top'):
        if status in ['no', 'fail', 'failed', 'false', False, 0]:
            status = 'error'
        elif status in ['yes', 'done', True, 1]:
            status = 'success'
        if status not in ['error', 'success']:
            status = 'normal'
        if position not in ['top', 'bottom']:
            position = 'top'
        message = message.replace("'", '')
        script = self.load('essage')+ "<script>var essage_message={message: '%s', status: '%s', placement: '%s'}; if(Essage){Essage.show(essage_message, %s)}</script>" %\
                                       (message, status, position, timeout)
        return script


    def show_note(self, *args, **kwargs):
        return self.show_notification(*args, **kwargs)


    def re_dom(self, html_content, rules):
        # 重新渲染 html 的某些 dom 结构
        # 一般用不到，但是在一些特定的或者封闭性的平台中或许有用，比如微信的公众号
        html_content = re_html_dom(html_content, rules)
        return html_content


    def linebreaks(self, content):
        content = smart_unicode(content)[:50000]
        return _linebreaks(content)

    @property
    def random_dom_id(self):
        # 返回一个随机的dom_id
        return 'd_%s' % uuid.uuid4().hex


    def get_dom_id(self, v=None):
        # 将一个value转为dom_id的格式
        return 'dd_%s' % get_md5(v)

    @staticmethod
    def join_url(url, *args, **kwargs):
        new_url = join_url(url, **kwargs)
        return new_url

    @staticmethod
    def get_url(url):
        return get_url_with_prefix(url)

    @staticmethod
    def url(url):
        # 同 get_url
        return get_url_with_prefix(url)


    @staticmethod
    def i18n(key, *args):
        key = smart_unicode(key)
        if not args:
            # 获取
            if key.startswith('_'): # 以_开头的，摘1后，返回原值
                return key[1:]
            lang = get_language()
            key1 = '%s-%s' % (key, lang)
            i18n_data = getattr(g, 'i18n', {})
            matched_value = i18n_data.get(key1) or i18n_data.get(key)
            return matched_value or key
        elif len(args)<=2:
            if not getattr(g, 'i18n', None):
                g.i18n = {}
            if len(args) == 2: # 指定了 lang 的
                value, lang = args
                lang = smart_unicode(lang).strip().lower()
                key = '%s-%s' % (key, lang)
            else: # 没有指定lang，设定默认值
                value = args[0]
            g.i18n[key] = value
            # 设置的
        # at last
        return ''


    @staticmethod
    def paginator(paginator=None, style='simple', **kwargs):
        # todo 页码分页的问题还要继续对应
        if 'simple' in kwargs and not kwargs.get('simple'):
            # 对旧的兼容，最开始的时候，auto 风格的调用是 simple=False
            style = 'auto'
        # 构建上下页的label，当然本身也是可以传入 HTML 片段的
        if 'pre_label' not in kwargs:
            if style == 'mini':
                pre_label = '<'
            else:
                pre_label = '&laquo; Prev'
            kwargs['pre_label'] = pre_label
        if 'next_label' not in kwargs:
            if style == 'mini':
                next_label = '>'
            else:
                next_label = 'Next &raquo;'
            kwargs['next_label'] = next_label

        paginator = paginator or get_paginator()
        if not paginator:
            return '' # ignore

        return render_api_template('paginator',
                                    paginator=paginator, paginator_style=style, return_html=True, **kwargs)



@cache_result
def h():
    return Html()

@cache_result
def html():
    return Html()


#coding: utf8
from __future__ import absolute_import
import os
from farbox_lite.compiler.utils import get_file_list
from .jade2jinja import compile_jade_file
from .css import compile_css_file


def is_file_to_compile(filepath):
    if os.path.isdir(filepath):
        return True
    elif os.path.isfile(filepath):
        ext = os.path.splitext(filepath)[-1].lower()
        if ext in ['.jade', '.less', '.scss']:
            return True
    # at last
    return False


def compile_site_templates(templates_root):
    filepaths = get_file_list(templates_root, filter_func=is_file_to_compile)
    for filepath in filepaths:
        ext = os.path.splitext(filepath)[-1].lower()
        if ext in ['.less', '.scss']:
            compile_css_file(filepath)
        elif ext == '.jade':
            compile_jade_file(filepath)



#coding: utf8
from __future__ import absolute_import
from dateutil.parser import parse as parse_date
import datetime, re, unidecode, time
from collections import OrderedDict
from farbox_lite.utils import smart_unicode, smart_str, LazyDict, get_md5, UnicodeWithAttrs
from farbox_lite.utils.functional import curry


def is_public(value):
    if value in [True, 'public', 'open', 'on', 'published', 'true', 'yes']:
        return True
    else:
        return False

is_on = is_public

def is_closed(value):
    if value in ['no', 'false', False, 'off', 'close']:
        return True
    else:
        return False


def dict_parts(dt, parts):
    return {part: dt.get(part) for part in parts}


def list_to_dict(keys, values):
    data = {}
    for i, value in enumerate(values):
        if isinstance(value, (str, unicode)):
            value = value.strip()
        try:
            key = keys[i]
            data[key] = value
        except:
            pass
    for key in keys:
        if key not in data:
            data[key] = None
    return data


def get_value_from_data(data, attr, default=None):
    if isinstance(data, dict) and attr in data:
        # dict 类型，直接返回value，attr本身可能包含 .
        return data[attr]
    try:
        attrs = attr.split('.')[:25] # 最多允许25个遍历
        for attr in attrs:

            # 因为系统默认是建立在 path 上的，所以字符串的 parent 会寻找 parent
            if attr == 'parent_path' and isinstance(data, (str, unicode)):
                if '/' not in data:
                    return ''
                else:
                    return data.rsplit('/', 1)[0]

            # 最后一个dt是真实的value
            if type(data) in [dict, OrderedDict]:
                data = data.get(attr, None)
            else:
                try:
                    data = getattr(data, attr, None)
                except: # jinja 中的 Undefined 会触发错误
                    return data
            if data is None or isinstance(data, LazyDict):  # 到底了
                if default is not None:
                    return default
                else:
                    return None
    except RuntimeError: # 一般是在外部调用g/request的时候会遇到
        return None
    return data

def set_attr_to_obj(obj, attr, value):
    try:
        setattr(obj, attr, value)
    except RuntimeError: # 一般是在外部调用g/request的时候会遇到
        pass


def merge_datas(*args):
    # 最后一个字典的值最优先，逻辑为依次覆盖
    data = dict()
    for dt in args:
        if type(dt) != dict: # 非字典
            data.update(getattr(data, '__dict__', {}))
        else:  # 字典
            data.update(dt)
    return data


def clone_field_from_other_dict(current, other, fields=()):
    # 一个dict，从另外的dict，按照指定的fields进行value clone
    for field in fields:
        if field in other:
            value_on_other = other.get(field)
            current[field] = value_on_other
    return current


class dict_to_obj(object):
    def __init__(self, dt):
        for a, b in dt.items():
            if isinstance(b, (list, tuple)):
               setattr(self, a, [self.__class__(x) if isinstance(x, dict) else x for x in b])
            else:
               setattr(self, a, self.__class__(b) if isinstance(b, dict) else b)


def to_number(value, default_if_fail=None, max_value=None, min_value=None, number_type_func=None):
    if isinstance(value, (str, unicode)):
        value = value.strip()
    if not value and type(value)!=int:
        return default_if_fail
    try:
        value = float(value)
        if number_type_func:
            value = number_type_func(value)
    except:
        value = default_if_fail
    if max_value is not None and value > max_value:
        value = max_value
    if min_value is not None and value < min_value:
        value = min_value
    return value


def to_float(value, default_if_fail=None, max_value=None, min_value=None):
    if isinstance(value, (str, unicode))  and '/' in value and value.count('/')==1: # 分数
        k1, k2 = value.split('/', 1)
        k1 = to_float(k1)
        k2 = to_float(k2)
        if k1 and k2:
            value = k1/k2
    return to_number(value, default_if_fail=default_if_fail, max_value=max_value, min_value=min_value)


to_int = curry(to_number, number_type_func=int)



def string_to_int(value):
    if not value:
        return value
    if isinstance(value, int):
        return value
    if isinstance(value, (str, unicode)): # 提取字符串中的整数部分，作为value
        re_s = re.search('\d+', value)
        if re_s:
            int_value = re_s.group()
            return to_int(int_value)
    # at last
    return to_int(value)


def is_str(text, includes=''):
    if not isinstance(text, (str, unicode)):
        return False
    text = text.strip()
    if includes:
        for s in includes: text=text.replace(s, '')
    if not text:
        return False
    return bool(re.match(r'[a-z0-9_\-]+$', text, flags=re.I))


def unique_list(data):
    # 类似 set 功能，但是保留原来的次序
    new_data = []
    for row in data:
        if row not in new_data:
            new_data.append(row)
    return new_data


def string_to_list(value):
    if not value:
        return []
    if isinstance(value, (str, unicode)) and value:
        if ',' in value:
            ls = value.split(',')
        elif u'，' in value:
            ls = value.split(u'，')
        else:
            ls = value.split(' ')
        ls = [item.strip() for item in ls if item]
        return unique_list(ls)
    elif type(value) in [list, tuple]:
        value = unique_list(value)
        return [smart_unicode(row).strip() for row in value if row]
    else:
        return [smart_str(value)]


def simple_kv_dict(data_list):
    # 将一个字符串的 list，内用空格或者冒号形成的 kv 结构转为 dict 模式
    if isinstance(data_list, (str, unicode)):
        data_list = string_to_list(data_list)
    if not isinstance(data_list, (list, tuple)):
        return data_list # ignore
    result_dict = {}
    for row in data_list:
        if not isinstance(row, (str, unicode)):
            continue
        parts = re.split(u'\uff1a|:', row, 1)
        if len(parts) == 2:
            k, v = parts
            result_dict[k.strip()] = v.strip()
        elif len(parts) == 1 and ' ' in row: # 尝试用空格分割
            parts = row.split(' ', 1)
            k, v = parts
            result_dict[k.strip()] = v.strip()
    return result_dict




def slugify(value, must_lower=True, auto=False):
    # auto=True 表示是自动生成的
    value= smart_unicode(value)
    value = unidecode.unidecode(value).strip()
    if must_lower:
        value = value.lower()
    value = re.sub(r'[ &~,"\':*+?#{}()<>\[\]]', '-', value).strip('-')  # 去掉非法的url字符
    value = re.sub(r'-+', '-', value)  # 替换连续的 --的或者-/
    value = value.replace('-/', '/')
    if auto: # 去掉可能的序号, 这样自动生成的 url 美观一点
        value = re.sub(r'^\d{1,2}(\.\d+)? ', '', value).strip() or value
        value = value.strip('-_')
    value = value.strip('/') # 头尾不能包含 /
    return value


def to_date(value):
    if not value:
        return None
    if isinstance(value, datetime.datetime):
        return value
    elif isinstance(getattr(value, 'core', None), datetime.datetime): # 被Date给wrap了
        return getattr(value, 'core', None)
    else:
        try:
            return parse_date(value)
        except:
            return None


def auto_type(value):
    # 自动类型，主要是 str 类的转为 number 的可能
    if not isinstance(value, (str, unicode)):
        return value
    value = value.strip()
    if re.match('^\d+$', value):
        return int(value)
    if re.match('^\d+\.(\d+)?$', value):
        return float(value)
    if re.match('^\d+/\d+$', value): # 分数
        new_value = to_float(value)
        if new_value:
            return new_value
    if value in ['True', 'true']:
        return True
    if value in ['False', 'false']:
        return False
    # at last
    return value


def re_sub(pattern, repl, string, count=0, flags=0):
    # 避免 re 的 sub 内如果 content 有 #/g 这种特定的非安全写法
    founds = re.findall(pattern, string, flags=flags)
    if count:
        founds = founds[:count]
    for found in founds:
        string = string.replace(found, repl, 1)
    return string


def simple_yaml_str(data, full=False):
    to_return = ''
    for key, value in data.items():
        to_return += '%s: %s\n' % (key, value)
    if full:
        to_return = '---\n%s---\n' % to_return
    return to_return


def count_words(content):
    # http://www.khngai.com/chinese/charmap/tbluni.php?page=0
        # 4e00 - 9fff是中文字符集的头尾 # 19968 - 40959
        # 3000 - 4db0 日文   # 3000-19888
        # 1100–11FF  # 韩文
        # 44032 - 55203 3130-12687 43360-43391  55216-55295
        # 0.5s 大概可以处理 176w的unicode
        # 2k字算的话，大概0.0005s，应该不会有性能问题
    content = smart_unicode(content)
    total_words = len(re.findall(ur'[\w\-_/]+|[\u1100-\ufde8]', content)) # 直接 1100 - 65000
    return total_words


def cut_content_by_words(content, max_words, mark=u'...'):
    if not isinstance(max_words, int):
        return content
    if max_words < 1:
        return content
    if max_words > 2000:
        max_words = 2000 # 最大不能超过这个，不然性能问题
    content = smart_unicode(content)
    iter_found = re.finditer(ur'[\w\-_/]+|[\u1100-\ufde8]', content)
    n = 0
    end = 0
    for i in iter_found:
        end = i.end()
        n += 1
        if n >= max_words:
            break
    if end:
        new_content = content[:end]
        if len(content) > end:
            mark = smart_unicode(mark)
            new_content += mark
            new_content = UnicodeWithAttrs(new_content)
            new_content.has_more = True
        return new_content
    else:
        return content



def str_equal(s1, s2, lower=False):
    if not s1 and not s2:  # 都为空的时候，返回False
        return False
    s1 = smart_unicode(s1)
    s2 = smart_unicode(s2)
    if lower:
        return s1.lower() == s2.lower()
    else:
        return s1 == s2


def bytes2human(num):
    if isinstance(num, (str, unicode)):
        return num
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'TB')

def seconds2human(seconds):
    mins = int(seconds/60)
    seconds_left = int(seconds - mins*60)
    return '%s:%02d' % (mins, seconds_left)


# 将一个整数，压码为字符串
def encode_int_to_str(n):
    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    alphabet_length = len(alphabet)
    encoded = ''
    while n > 0:
        n, r = divmod(n, alphabet_length)
        encoded = alphabet[r]+encoded
    return encoded


def decode_str_to_int(s):
    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    alphabet_length = len(alphabet)
    s_length = len(s)
    d = dict((ch, i) for (i, ch) in enumerate(alphabet))
    n = sum(d[ch] * alphabet_length**(s_length-i-1) for (i,ch) in enumerate(s))
    return n


# 提取附属星系， 比如 hello (x=y, k=v) 其中括号内的内容会形成 kv 结构的字典, 仅限单行
def extract_extra_info(line):
    line = line.replace('&quot;', '"').replace('&apos;', "'")
    line = line.replace('<em>', '_').replace('</em>', '_') # 跟Markdown解析后的HTML混淆了，转义回来
    extra_info ={}
    if isinstance(line, (str, unicode)) and '\n' not in line and '{' in line and line.endswith('}'): # 对 {}的兼容
        # 因为要提取的内部，是不能包含 () 的， 所以用 {} 进行替换； 但前提是 尾部由 } 结尾，以免造成不必要的处理
        line = line.replace('{', '(').replace('}', ')')
    if isinstance(line, (str, unicode)) and '\n' not in line and '(' in line and line.endswith(')'):
        if ',' not in line and '&' in line:
            line = line.replace('&', ',')

        c = re.search(r'(.*?)\((.*?)\)$', line)
        if c:
            new_line, extra_info_s = c.groups()
            line = new_line.strip()
            # lazy_list(type=dom_list, form_keys=[key, value], title='Content')  -> 这个分割比较困难
            if ', ' in extra_info_s: # 为了允许 , 也能用，就要确保field之间的声明用 “, ”，确保后续的空格
                parts = extra_info_s.split(', ')
            else:
                parts = extra_info_s.split(',')
            for extra_info_part in parts:
                if '=' not in extra_info_part:
                    continue
                k, v = extra_info_part.split('=', 1)
                k = k.strip()
                v = v.strip('\'"').strip()
                if re.match('\d+$', v):
                    v = int(v)
                elif re.match(r'\d+\.\d+$', v):
                    v = float(v)
                elif (v.startswith('[') and v.endswith(']')) or (v.startswith('(') and v.endswith(')')): # list
                    v = v[1:-1]
                    v = string_to_list(v)
                elif v.lower() in ['true', 'yes']:
                    v = True
                elif v.lower() in ['false', 'no']:
                    v = False
                extra_info[k] = v
    return line, extra_info



def make_list_small(ls, keys):
    result = []
    for row in ls:
        if not isinstance(row, dict):
            result.append(row)
        else:
            new_value = {key: row.get(key) for key in keys}
            result.append(new_value)
    return result



def all_has(doc, fields):
    if not doc:
        return False
    for field in fields:
        if not doc.get(field):
            return False
    return True



class cached(object):
    """
    @cached
    def xxxx():
        xxxx

    @cached(60, cache_key=None or other...)
    def xxx():
        xxxxxxx
    """
    def __init__(self, first_arg=None, cache_key=None):
        self.cached_function_responses = {}
        self.direct_func = None
        self.default_cache_key = cache_key

        if hasattr(first_arg, '__call__'):
            # 直接调用了这是
            # @cached
            # def xxxx(): .....
            self.max_age = 0
            self.direct_func = first_arg
        else:
            # 指定了缓存的时间
            # @cached(60)
            # def xxxx(): .......
            if not isinstance(first_arg, int):
                self.max_age = 0
            else:
                self.max_age = first_arg or 0

    def __call__(self, *args, **kwargs):
        def _func(*func_args, **func_kwargs):
            now = time.time()

            if self.direct_func:
                func = self.direct_func
            else:
                func = args[0]

            # 计算缓存的key值，根据传入的函数参数不同可以推断，如果没有函数传入，则相当于func本身的id（整数）
            # func_id 是作为cache_key的base
            if self.default_cache_key:
                base_cache_key = self.default_cache_key
            else:
                base_cache_key = id(func)
            if func_args or func_kwargs:
                # 可能第一个变量传入的是 self 这种实例对象
                func_args_s_list = [] # 形成缓存key用的一个list
                for func_arg in func_args:
                    if isinstance(func_arg, dict) and func_arg.get('_id'):
                        func_args_s_list.append(smart_unicode(func_arg.get('_id')))
                    elif not isinstance(func_arg, (str, unicode, dict, tuple, list, bool, int, float)):
                        func_args_s_list.append(str(id(func_arg)))
                    else:
                        func_args_s_list.append(smart_unicode(func_arg))
                var_key = smart_unicode('-'.join(func_args_s_list)) + smart_unicode(func_kwargs)
                var_md5 = get_md5(var_key)
                cache_key = '%s-%s' % (base_cache_key, var_md5)
            else:
                cache_key = base_cache_key

            if cache_key in self.cached_function_responses:
                cached_obj = self.cached_function_responses[cache_key]
                cached_data = cached_obj['data']
                cached_time = cached_obj['cache_time']
                if not self.max_age: # 相当于直接缓存，没有过期时间
                    return cached_data
                else: # 查看是否过期
                    if now - cached_time < self.max_age:
                        return cached_data
            result = func(*func_args, **func_kwargs)
            self.cached_function_responses[cache_key] = dict(data=result, cache_time=now)
            return result

        if self.direct_func:
            return _func(*args, **kwargs)
        else:
            return _func


def split_list(ls, size_per_part):
    for i in range(0, len(ls), size_per_part):
        yield ls[i:i + size_per_part]


def auto_choose(key, choices):
    # 一个key过来，会指定一个choice，而不是随机的; choices 本身要是已经sort的list。
    if not choices:
        return # ignore, return None
    if len(choices) == 1:
        return choices[0]
    if isinstance(key, dict) and '_id' in key:
        # 传入的key可能是一个数据库对象，以其_id作为key
        key = key['_id']
    key = get_md5(key)
    number = int(''.join([str(ord(s)) for s in key[:5]]))
    key_number = number % len(choices)
    return choices[key_number]


# coding: utf8
from __future__ import absolute_import
import datetime
from farbox_lite.template_system.functions.attr_patches.models.date import Date

def date(obj):
    if isinstance(obj, datetime.datetime):
        return Date(obj)
    else:
        return obj



#coding: utf8
from __future__ import absolute_import
import os
import unidecode
import re
import urllib
import datetime
from dateutil.parser import parse as parse_date
from farbox_markdown.compile_md import compile_markdown
from farbox_lite.utils import smart_unicode, smart_str, get_relative_path, string_to_list
from farbox_lite.compiler.utils import get_sort_and_name_from_filepath


def count_words(content):
    content = smart_unicode(content)
    total_words = len(re.findall(ur'[\w\-_/]+|[\u1100-\ufde8]', content))
    return total_words

def slugify(value, must_lower=True, auto=False):
    # auto=True 表示是自动生成的
    value= smart_unicode(value)
    value = unidecode.unidecode(value).strip()
    if must_lower:
        value = value.lower()
    value = re.sub(r'[ &~,"\':*+?#{}()<>\[\]]', '-', value).strip('-')  # 去掉非法的url字符
    value = re.sub(r'-+', '-', value)  # 替换连续的 --的或者-/
    value = value.replace('-/', '/')
    if auto: # 去掉可能的序号, 这样自动生成的 url 美观一点
        value = re.sub(r'^\d{1,2}(\.\d+)? ', '', value).strip() or value
        value = value.strip('-_')
    value = value.strip('/') # 头尾不能包含 /
    return value


def to_date(value):
    if not value:
        return None
    if isinstance(value, datetime.datetime):
        return value
    elif isinstance(getattr(value, 'core', None), datetime.datetime): # 被Date给wrap了
        return getattr(value, 'core', None)
    else:
        try:
            return parse_date(value)
        except:
            return None


def get_images_from_html(content, relative_to_site=False):
    raw_images = re.findall("""<\s*img.*?src=['"]([^'"]+).*?>""", content, flags=re.I)
    images = []
    for image_src in raw_images:
        if image_src in images:
            continue
        if relative_to_site and ('://' in image_src or image_src.startswith('//')):
            continue
        images.append(image_src)
    return images



def get_md_meta_value(compiled_content, key_name, default=None):
    metadata = getattr(compiled_content, 'metadata', {})
    if not isinstance(metadata, dict):
        metadata = {}
    if key_name in metadata:
        value = metadata.get(key_name)
        if value is None:
            value = default
    else:
        value = default
    if isinstance(value, (str, unicode)):
        value = value.strip()
    # 根据default的值，自动格式化得到的value
    if default is not None and not isinstance(value, basestring) and value is not None:
        value = type(default)(value)  # 变量格式化
    return value


def filepath_to_post_doc(filepath, root=None):
    if not os.path.isfile(filepath):
        return {}
    if not root:
        root = os.path.dirname(filepath)
    filepath = smart_unicode(filepath)
    relative_path = get_relative_path(filepath, root=root)
    relative_path = smart_unicode(relative_path)
    filename = os.path.split(filepath)[-1]
    just_name = os.path.splitext(filename)[0]
    slash_number = relative_path.count('/')

    with open(filepath, 'rb') as f:
        raw_content = f.read()
    raw_content = smart_unicode(raw_content)
    compiled_content = compile_markdown(raw_content)

    sort, name = get_sort_and_name_from_filepath(filepath)
    title = get_md_meta_value(compiled_content, 'title', '')
    #文中没有声明，看有没有唯一的H1作为title
    if not title and getattr(compiled_content, 'title', None) and getattr(compiled_content, 'title_from_post', False):
        title = compiled_content.title
    if not title: # 仍然没有title的，就取文件名
        title = just_name
        if title == 'index': # index doc, get parent folder's name
            parent_folder = os.path.dirname(filepath)
            title = os.path.split(parent_folder)[-1]
    title = smart_unicode(title.strip())

    tags = get_md_meta_value(compiled_content, 'tags') or get_md_meta_value(compiled_content, 'tag')
    tags = string_to_list(tags)
    if not isinstance(tags, (list, tuple)):
        tags = []

    if relative_path in ['robots.txt', 'robot.txt']:
        status = 'system'
    elif re.match(r'^_[a-z]{1,20}/', relative_path, re.I) and not re.match(r'_posts?/', relative_path)\
            and not get_md_meta_value(compiled_content,'status'):
        # 位于_xxx 目录下的，默认为status 为 xxx (如果没有声明的话)
        # 但是对_post/_posts下的目录不处理status，兼容Jekyll
        status = relative_path.split('/')[0].lower().strip('_')
    else:
        status = get_md_meta_value(compiled_content, 'status', 'public') or 'public'
    status = status.strip().lower()

    # url_path是全小写
    url_path = get_md_meta_value(compiled_content, 'url', '') or get_md_meta_value(compiled_content, 'url_path', '')
    if url_path and not isinstance(url_path, basestring):
        url_path = smart_unicode(url_path)
    if not url_path: # 如果是用户自己指定的url，则不管对错，都保存; 反之则是通过relative_path解析一个url
        url_path = relative_path.rsplit('.', 1)[0]
        url_path = slugify(url_path, auto=True).lower() # 不能以slash开头,并且确保小写
    else: # 用户自己声明的 url，不做大小写处理，保留原始状态
        url_path = slugify(url_path, must_lower=False) # 替代掉不适合做url的字符
    if '%' in url_path:
        # 被编码的url，特别是wordpress转过来的
        _url_path = urllib.unquote(smart_str(url_path))
        if url_path != _url_path:
            url_path = smart_unicode(_url_path)
    url_path = url_path.lstrip('/')
    url_path = re.sub(r'/{2,}', '/', url_path)
    url_path = url_path or '--' # 可能存在空的情况...


    # date
    # 文件，client属性中的最后修改时间
    updated_at = datetime.datetime.fromtimestamp(os.path.getmtime(filepath))
    try:
        updated_at_s = smart_str(get_md_meta_value(compiled_content, 'date', ''))
        if updated_at_s:
            updated_at_s = updated_at_s.strip()
        if updated_at_s and re.match('\d{4}\.\d+\.\d+$', updated_at_s):
            # # 2018.3.19  这种日期形式的 转为 xxxx-xx-xx
            updated_at_s = updated_at_s.replace('.', '-')
        if not updated_at_s:
            # 兼容 '2012-12-12 12-12 or 2012-12-12 12-12-12 这种格式'
            done = False
            if re.match(r'^\d+-\d+-\d+ \d+-\d+(-\d+)?$', just_name):
                part1, part2 = just_name.split(' ', 1)
                try:
                    updated_at = parse_date('%s %s' % (part1, part2.replace('-', ':')))
                    done = True
                except:
                    pass
            if not done:
                # 尝试从文件名中获取 2012-1?2-1?2, date模式
                date_search = re.search('/?(\d{4}-\d{1,2}-\d{1,2})[^/]*', relative_path)
                if date_search: # 可以从文件的路径中取， 兼容jekyll
                    updated_at_s = date_search.groups()[0]
        if updated_at_s:
            try:
                updated_at = parse_date(updated_at_s)
            except:
                pass
    except (ValueError, TypeError):
        pass


    metadata = compiled_content.metadata.copy()
    doc = dict(
        title = title,
        status = status,
        tags = tags,
        url_path = url_path,
        date = updated_at,

        cover = compiled_content.cover,
        toc = compiled_content.toc,
        metadata = metadata,
        images = get_images_from_html(compiled_content) or [],

        raw_content = raw_content,
        content = compiled_content,
        text_length = len(raw_content),
        text_words = count_words(raw_content),

        slash_number = slash_number,

        path = relative_path.lower(),
        filepath = filepath,
        md_filepath = filepath,
        mtime = os.path.getmtime(filepath),
        sort = sort,
        position = sort,

        type = 'post',

        id = filepath,

        url = '/post/%s' % url_path
    )

    return doc






# coding: utf8
from __future__ import absolute_import

import re, random, os
from flask import g, request
from jinja2.runtime import Undefined

from farbox_lite.utils.functional import curry, cached_property
from farbox_lite.utils import smart_unicode
from farbox_lite.utils.lazy import get_value_from_data, to_int, string_to_int
from farbox_lite.utils.mime import guess_type

from farbox_lite.utils.web.path import get_request_offset_path_without_prefix
from farbox_lite.utils.web.cache_for_function import cache_result

from farbox_lite.render.apis.data import get_data, get_paginator, has_doc, get_doc

from farbox_lite.template_system.functions.namespace.utils.doc_list_utils import DocList

_max = max



class Image(object):
    # 一个具体图片的文档对象 wrap 之后的 data obj
    def __init__(self, doc):
        if not isinstance(doc, dict):
            doc = {}
        if doc and not doc.get('type')=='image': # 必须保证是图片类型
            doc = None

        self.md5 = doc.get('md5') or doc.get('path') # md5值
        self.id = doc.get('id') or doc.get('_id')
        self.raw = doc or {}

        self.is_image = True # 在某些场合，可以用于判断当前是否 Image 对象

    def __nonzero__(self):
        return bool(self.raw)


    def __getattr__(self, item):
        if item in self.raw:
            return self.raw.get(item)
        return self.__dict__.get(item, Undefined())

    @cached_property
    def length(self):
        return len(self.list_obj)


    @cached_property
    def name(self):
        path = self.raw.get('path') or ''
        pathname = os.path.split(path)[-1]
        return pathname


    @cached_property
    def url(self):
        return '/' + self.raw['path'].strip('/').lower()

    @cached_property
    def cover(self):
        return self.url




class Images(DocList):
    def __init__(self):
        DocList.__init__(self, 'images')

    def __getattr__(self, item):
        # 找不到attribute的时候，会调用getattr
        if re.match('^recent_?\d+$', item): # 获得最近的几篇文章
            limit = re.search('\d+', item).group()
            limit = to_int(limit)
            return self.get_recent(limit)

    def __getitem__(self, item):
        if hasattr(self.list_obj, '__getitem__'):
            return self.list_obj.__getitem__(item)

    def __nonzero__(self):
        return bool(self.list_obj)

    def __iter__(self):
        # 返回一个迭代器，用于 for 的命令
        obj = self.list_obj
        self.pager_name = 'images'
        if hasattr(obj, '__iter__'):
            return obj.__iter__()
        return obj


    def return_image_objs(self, docs):
        if not docs:
            return []
        objs = [Image(doc) for doc in docs]
        return objs



    def get_recent(self, limit=8):
        limit = to_int(limit, 8)
        limit = min(100, limit)
        image_docs = get_data(type='image', limit=limit, with_page=False, sort='-date')
        return self.return_image_objs(image_docs)


    def get_recent_images(self, limit=8):
        return self.get_recent(limit)


    def get_one(self, path):
        return self.find_one(path)


    @property
    def pager(self):
        # 由于获得分页的数据对象
        # 比如可以直接调用  posts.pager....
        return get_paginator(self.pager_name)

    @property
    def paginator(self):
        return self.pager


    @property
    def all(self):
        self.pager_name = 'all_images'
        return self._all


    @cached_property
    def _all(self):
        # 所有日志
        image_docs = get_data(types='image', pager_name='all_images', min_limit=self.min_per_page)
        return self.return_image_objs(image_docs)


    @cached_property
    def image(self):
        # 得到当前 url 下对应的 图片
        url_path = get_request_offset_path_without_prefix(offset=1)
        image_doc = self.get_one(url_path)
        if image_doc:
            g.file_doc = image_doc
        return Image(image_doc)


    @cached_property
    def list_obj(self):
        pager_name = 'images'
        # 针对默认 的 url 规则，自动对应默认的 posts 的调用
        if self.keywords: # 关键词搜索放在最前面
            return get_data(type='image', keywords=request.args.get('s'), limit=50, pager_name=pager_name, min_limit=self.min_per_page, sort=self.sort)
        if self.under is not None: # 指定了 under
            return get_data(type='image', path=self.under, pager_name=pager_name, min_limit=self.min_per_page, sort=self.sort)
        if self.request_path.startswith('/category/'): # 指定目录下的
            return get_data(type='image', path=get_request_offset_path_without_prefix(1), pager_name=pager_name, min_limit=self.min_per_page, sort=self.sort)
        # at last
        _images = get_data(type='image', pager_name=pager_name, min_limit=self.min_per_page, sort=self.sort)
        return self.return_image_objs(_images)

    @cached_property
    def counts(self):
        self_list_obj = self.list_obj # 先行调用
        if self.pager:
            return self.pager.total_count
        else:
            return 0

    @cached_property
    def images_count(self):
        return self.counts


@cache_result
def images():
    return Images()


# coding: utf8
from __future__ import absolute_import

from farbox_lite.render.apis.data import get_paginator  #, get_data as _get_data

from farbox_lite.template_system.functions.namespace.html import Html
from farbox_lite.template_system.functions.namespace.data import Data
from farbox_lite.utils.web.cache_for_function import cache_result
from .post import posts
from .image import images


@cache_result
def paginator():
    return get_paginator()

@cache_result
def pager():
    return get_paginator()


@cache_result
def post():
    _posts = posts()
    return _posts.post


def get_post():
    return post()


@cache_result
def image():
    _images = images()
    return _images.image



################# functions starts ##############

def load(source):
    return Html.load(source)


@cache_result
def get_data(*args, **kwargs):
    return Data.get_data(*args, **kwargs)



@cache_result
def _(key, *args): # i18n 专用函数名
    return Html.i18n(key, *args)


@cache_result
def i18n(key, *args): # i18n 专用函数名
    return Html.i18n(key, *args)





#coding: utf8
import os

MIMETYPES = {'.zirz': 'application/vnd.zul', '.conf': 'text/plain', '.obd': 'application/x-msbinder',
             '.obj': 'application/octet-stream', '.dtshd': 'audio/vnd.dts.hd', '.silo': 'model/mesh',
             '.cryptonote': 'application/vnd.rig.cryptonote', '.pml': 'application/vnd.ctc-posml',
             '.f4v': 'video/x-f4v', '.iso': 'application/octet-stream', '.hvs': 'application/vnd.yamaha.hv-script',
             '.ras': 'image/x-cmu-raster', '.rar': 'application/x-rar-compressed', '.mrc': 'application/marc',
             '.sdkd': 'application/vnd.solent.sdkm+xml', '.bcpio': 'application/x-bcpio', '.asf': 'video/x-ms-asf',
             '.sdkm': 'application/vnd.solent.sdkm+xml', '.list3820': 'application/vnd.ibm.modcap',
             '.ecelp7470': 'audio/vnd.nuera.ecelp7470', '.fh5': 'image/x-freehand', '.bz': 'application/x-bzip',
             '.onetoc2': 'application/onenote', '.ait': 'application/vnd.dvb.ait', '.onetmp': 'application/onenote',
             '.bmp': 'image/bmp', '.geo': 'application/vnd.dynageo',
             '.air': 'application/vnd.adobe.air-application-installer-package+zip', '.bmi': 'application/vnd.bmi',
             '.jad': 'text/vnd.sun.j2me.app-descriptor', '.xhvml': 'application/xv+xml', '.dvi': 'application/x-dvi',
             '.aif': 'audio/x-aiff', '.eps': 'application/postscript', '.hpid': 'application/vnd.hp-hpid',
             '.ott': 'application/vnd.oasis.opendocument.text-template', '.rss': 'application/rss+xml',
             '.ots': 'application/vnd.oasis.opendocument.spreadsheet-template', '.hlp': 'application/winhlp',
             '.cst': 'application/x-director', '.p10': 'application/pkcs10', '.csv': 'text/csv',
             '.p12': 'application/x-pkcs12', '.csp': 'application/vnd.commonspace',
             '.sus': 'application/vnd.sus-calendar', '.css': 'text/css', '.npx': 'image/vnd.net-fpx',
             '.otg': 'application/vnd.oasis.opendocument.graphics-template', '.otf': 'application/x-font-otf',
             '.csh': 'application/x-csh', '.otc': 'application/vnd.oasis.opendocument.chart-template',
             '.pdb': 'application/vnd.palm', '.clkp': 'application/vnd.crick.clicker.palette',
             '.pdf': 'application/pdf', '.hal': 'application/vnd.hal+xml', '.pl': 'text/plain',
             '.atomsvc': 'application/atomsvc+xml', '.xul': 'application/vnd.mozilla.xul+xml',
             '.chm': 'application/vnd.ms-htmlhelp', '.djv': 'image/vnd.djvu', '.py': 'text/x-python',
             '.ttf': 'application/x-font-ttf', '.xml': 'application/xml', '.ps': 'application/postscript',
             '.htm': 'text/html', '.m2a': 'audio/mpeg', '.fig': 'application/x-xfig',
             '.sig': 'application/pgp-signature', '.cab': 'application/vnd.ms-cab-compressed',
             '.tsv': 'text/tab-separated-values', '.uvp': 'video/vnd.dece.pd', '.xlm': 'application/vnd.ms-excel',
             '.ltf': 'application/vnd.frogans.ltf', '.xlb': 'application/vnd.ms-excel',
             '.prc': 'application/x-mobipocket-ebook', '.pre': 'application/vnd.lotus-freelance',
             '.prf': 'application/pics-rules', '.car': 'application/vnd.curl.car',
             '.tsd': 'application/timestamped-data', '.dd2': 'application/vnd.oma.dd2+xml',
             '.cat': 'application/vnd.ms-pki.seccat', '.dms': 'application/octet-stream',
             '.sc': 'application/vnd.ibm.secure-container', '.ief': 'image/ief', '.mp4s': 'application/mp4',
             '.uvs': 'video/vnd.dece.sd', '.jpgm': 'video/jpm', '.c4d': 'application/vnd.clonk.c4group',
             '.c4g': 'application/vnd.clonk.c4group', '.c4f': 'application/vnd.clonk.c4group',
             '.texinfo': 'application/x-texinfo', '.p8': 'application/pkcs8', '.dmg': 'application/octet-stream',
             '.c4p': 'application/vnd.clonk.c4group', '.c4u': 'application/vnd.clonk.c4group',
             '.vis': 'application/vnd.visionary', '.viv': 'video/vnd.vivo', '.ddd': 'application/vnd.fujixerox.ddd',
             '.link66': 'application/vnd.route66.link66+xml', '.tmo': 'application/vnd.tmobile-livetv',
             '.ext': 'application/vnd.novadigm.ext', '.exi': 'application/exi', '.csml': 'chemical/x-csml',
             '.mus': 'application/vnd.musician', '.opf': 'application/oebps-package+xml',
             '.exe': 'application/x-msdownload', '.xpw': 'application/vnd.intercon.formnet',
             '.xpr': 'application/vnd.is-xpr', '.xps': 'application/vnd.ms-xpsdocument', '.dsc': 'text/prs.lines.tag',
             '.xpx': 'application/vnd.intercon.formnet', '.mscml': 'application/mediaservercontrol+xml',
             '.rep': 'application/vnd.businessobjects', '.xpm': 'image/x-xpixmap', '.mpeg': 'video/mpeg',
             '.ms': 'text/troff', '.m21': 'application/mp21', '.sisx': 'application/vnd.symbian.install',
             '.m4v': 'video/x-m4v', '.spq': 'application/scvp-vp-request', '.fxpl': 'application/vnd.adobe.fxp',
             '.ami': 'application/vnd.amiga.ami', '.fm': 'application/vnd.framemaker', '.ram': 'audio/x-pn-realaudio',
             '.dssc': 'application/dssc+der', '.fh': 'image/x-freehand', '.sgml': 'text/sgml',
             '.spf': 'application/vnd.yamaha.smaf-phrase', '.xlt': 'application/vnd.ms-excel',
             '.spc': 'application/x-pkcs7-certificates', '.spl': 'application/x-futuresplash',
             '.bat': 'application/x-msdownload', '.mgp': 'application/vnd.osgeo.mapguide.package',
             '.emma': 'application/emma+xml', '.eml': 'message/rfc822', '.lvp': 'audio/vnd.lucent.voice',
             '.fhc': 'image/x-freehand', '.gac': 'application/vnd.groove-account', '.cww': 'application/prs.cww',
             '.efif': 'application/vnd.picsel', '.yin': 'application/yin+xml', '.wad': 'application/x-doom',
             '.saf': 'application/vnd.yamaha.smaf-audio', '.txf': 'application/vnd.mobius.txf',
             '.utz': 'application/vnd.uiq.theme', '.txd': 'application/vnd.genomatix.tuxedo', '.m2v': 'video/mpeg',
             '.wrl': 'model/vrml', '.wav': 'audio/x-wav', '.rsd': 'application/rsd+xml', '.mxml': 'application/xv+xml',
             '.xsm': 'application/vnd.syncml+xml', '.xbm': 'image/x-xbitmap', '.txt': 'text/plain',
             '.jlt': 'application/vnd.hp-jlyt', '.xbd': 'application/vnd.fujixerox.docuworks.binder',
             '.wax': 'audio/x-ms-wax', '.mlp': 'application/vnd.dolby.mlp', '.qxd': 'application/vnd.quark.quarkxpress',
             '.twd': 'application/vnd.simtech-mindmapper', '.dna': 'application/vnd.dna', '.ts': 'video/mp2t',
             '.tr': 'text/troff', '.distz': 'application/octet-stream', '.ahead': 'application/vnd.ahead.space',
             '.smil': 'application/smil+xml', '.fnc': 'application/vnd.frogans.fnc',
             '.et3': 'application/vnd.eszigno3+xml', '.xif': 'image/vnd.xiff', '.dae': 'model/vnd.collada+xml',
             '.daf': 'application/vnd.mobius.daf', '.cer': 'application/pkix-cert',
             '.smf': 'application/vnd.stardivision.math', '.ufd': 'application/vnd.ufdl',
             '.dotm': 'application/vnd.ms-word.template.macroenabled.12', '.smi': 'application/smil+xml',
             '.uu': 'text/x-uuencode', '.3gp': 'video/3gpp', '.pub': 'application/x-mspublisher',
             '.gxt': 'application/vnd.geonext', '.jisp': 'application/vnd.jisp', '.sfs': 'application/vnd.spotfire.sfs',
             '.imp': 'application/vnd.accpac.simply.imp', '.mpn': 'application/vnd.mophun.application',
             '.etx': 'text/x-setext', '.jpgv': 'video/jpeg', '.iges': 'model/iges', '.ogv': 'video/ogg',
             '.movie': 'video/x-sgi-movie', '.dxp': 'application/vnd.spotfire.dxp', '.png': 'image/png',
             '.mhtml': 'message/rfc822', '.tar': 'application/x-tar', '.pnm': 'image/x-portable-anymap',
             '.pnt': 'image/x-macpaint', '.mqy': 'application/vnd.mobius.mqy', '.asm': 'text/x-asm',
             '.dxf': 'image/vnd.dxf', '.rnc': 'application/relax-ng-compact-syntax',
             '.tao': 'application/vnd.tao.intent-module-archive',
             '.xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet', '.wvx': 'video/x-ms-wvx',
             '.wgt': 'application/widget', '.hpgl': 'application/vnd.hp-hpgl', '.ktx': 'image/ktx',
             '.dotx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.template',
             '.ecelp9600': 'audio/vnd.nuera.ecelp9600', '.cu': 'application/cu-seeme',
             '.chrt': 'application/vnd.kde.kchart', '.wsdl': 'application/wsdl+xml', '.dwg': 'image/vnd.dwg',
             '.dwf': 'model/vnd.dwf', '.fg5': 'application/vnd.fujitsu.oasysgp', '.dcurl': 'text/vnd.curl.dcurl',
             '.gex': 'application/vnd.geometry-explorer', '.ei6': 'application/vnd.pg.osasli',
             '.xlsb': 'application/vnd.ms-excel.sheet.binary.macroenabled.12', '.mcd': 'application/vnd.mcd',
             '.odft': 'application/vnd.oasis.opendocument.formula-template',
             '.nnd': 'application/vnd.noblenet-directory', '.pgn': 'application/x-chess-pgn',
             '.mwf': 'application/vnd.mfer', '.str': 'application/vnd.pg.format', '.g2w': 'application/vnd.geoplan',
             '.stw': 'application/vnd.sun.xml.writer.template', '.sti': 'application/vnd.sun.xml.impress.template',
             '.nns': 'application/vnd.noblenet-sealer', '.stk': 'application/hyperstudio',
             '.crl': 'application/pkix-crl', '.ipfix': 'application/ipfix', '.stl': 'application/vnd.ms-pki.stl',
             '.semf': 'application/vnd.semf', '.bed': 'application/vnd.realvnc.bed',
             '.stc': 'application/vnd.sun.xml.calc.template', '.crd': 'application/x-mscardfile',
             '.std': 'application/vnd.sun.xml.draw.template', '.m1v': 'video/mpeg', '.stf': 'application/vnd.wt.stf',
             '.gmx': 'application/vnd.gmx', '.ifb': 'text/calendar', '.avi': 'video/x-msvideo',
             '.fgd': 'application/x-director', '.dp': 'application/vnd.osgi.dp', '.mrcx': 'application/marcxml+xml',
             '.cla': 'application/vnd.claymore', '.pwn': 'application/vnd.3m.post-it-notes', '.wml': 'text/vnd.wap.wml',
             '.mht': 'message/rfc822', '.pskcxml': 'application/pskc+xml', '.wma': 'audio/x-ms-wma',
             '.wmf': 'application/x-msmetafile', '.wmd': 'application/x-ms-wmd', '.mcurl': 'text/vnd.curl.mcurl',
             '.wmz': 'application/x-ms-wmz', '.clp': 'application/x-msclip', '.wmx': 'video/x-ms-wmx',
             '.rp9': 'application/vnd.cloanto.rp9', '.mc1': 'application/vnd.medcalcdata',
             '.m13': 'application/x-msmediaview', '.m14': 'application/x-msmediaview',
             '.dist': 'application/octet-stream', '.mp4a': 'audio/mp4', '.nml': 'application/vnd.enliven',
             '.kpt': 'application/vnd.kde.kpresenter', '.uvg': 'image/vnd.dece.graphic',
             '.kpr': 'application/vnd.kde.kpresenter', '.dra': 'audio/vnd.dra', '.g3w': 'application/vnd.geospace',
             '.oa2': 'application/vnd.fujitsu.oasys2', '.oa3': 'application/vnd.fujitsu.oasys3',
             '.gtar': 'application/x-gtar', '.p7m': 'application/pkcs7-mime', '.nws': 'message/rfc822',
             '.deb': 'application/x-debian-package', '.p7c': 'application/pkcs7-mime',
             '.p7b': 'application/x-pkcs7-certificates', '.def': 'text/plain', '.jp2': 'image/jp2',
             '.mts': 'model/vnd.mts', '.bpk': 'application/octet-stream', '.ez': 'application/andrew-inset',
             '.der': 'application/x-x509-ca-cert', '.p7s': 'application/pkcs7-signature',
             '.p7r': 'application/x-pkcs7-certreqresp', '.pntg': 'image/x-macpaint', '.ico': 'image/x-icon',
             '.fxp': 'application/vnd.adobe.fxp', '.acc': 'application/vnd.americandynamics.acc',
             '.bdf': 'application/x-font-bdf', '.c11amz': 'application/vnd.cluetrust.cartomobile-config-pkg',
             '.ace': 'application/x-ace-compressed', '.rmi': 'audio/midi', '.listafp': 'application/vnd.ibm.modcap',
             '.dfac': 'application/vnd.dreamfactory', '.acu': 'application/vnd.acucobol',
             '.wmlsc': 'application/vnd.wap.wmlscriptc', '.oas': 'application/vnd.fujitsu.oasys',
             '.c11amc': 'application/vnd.cluetrust.cartomobile-config', '.tex': 'application/x-tex',
             '.wri': 'application/x-mswrite', '.mesh': 'model/mesh', '.irp': 'application/vnd.irepository.package+xml',
             '.ecma': 'application/ecmascript', '.sse': 'application/vnd.kodak-descriptor',
             '.ssf': 'application/vnd.epson.ssf', '.sitx': 'application/x-stuffitx',
             '.oth': 'application/vnd.oasis.opendocument.text-web', '.tei': 'application/tei+xml',
             '.rtx': 'text/richtext', '.qtif': 'image/x-quicktime', '.uvv': 'video/vnd.dece.video',
             '.meta4': 'application/metalink4+xml', '.irm': 'application/vnd.ibm.rights-management',
             '.ice': 'x-conference/x-cooltalk', '.joda': 'application/vnd.joost.joda-archive',
             '.rpst': 'application/vnd.nokia.radio-preset', '.fdf': 'application/vnd.fdf',
             '.elc': 'application/octet-stream', '.eot': 'application/vnd.ms-fontobject',
             '.rpss': 'application/vnd.nokia.radio-presets', '.hps': 'application/vnd.hp-hps',
             '.mfm': 'application/vnd.mfmp', '.wbmp': 'image/vnd.wap.wbmp', '.in': 'text/plain',
             '.pas': 'text/x-pascal', '.sub': 'image/vnd.dvb.subtitle', '.paw': 'application/vnd.pawaafile',
             '.fbs': 'image/vnd.fastbidsheet', '.lrf': 'application/octet-stream',
             '.gim': 'application/vnd.groove-identity-message', '.kfo': 'application/vnd.kde.kformula',
             '.ims': 'application/vnd.ms-ims', '.gif': 'image/gif', '.lrm': 'application/vnd.ms-lrm',
             '.cdkey': 'application/vnd.mediastation.cdkey', '.atomcat': 'application/atomcat+xml',
             '.qwd': 'application/vnd.quark.quarkxpress', '.i2g': 'application/vnd.intergeo', '.rtf': 'application/rtf',
             '.mov': 'video/quicktime', '.twds': 'application/vnd.simtech-mindmapper', '.dxr': 'application/x-director',
             '.azw': 'application/vnd.amazon.ebook', '.qxt': 'application/vnd.quark.quarkxpress',
             '.azs': 'application/vnd.airzip.filesecure.azs', '.iif': 'application/vnd.shana.informed.interchange',
             '.sv4cpio': 'application/x-sv4cpio', '.ngdat': 'application/vnd.nokia.n-gage.data',
             '.azf': 'application/vnd.airzip.filesecure.azf', '.com': 'application/x-msdownload',
             '.wbxml': 'application/vnd.wap.wbxml', '.hh': 'text/x-c', '.mathml': 'application/mathml+xml',
             '.rmp': 'audio/x-pn-realaudio-plugin', '.wiz': 'application/msword', '.vcd': 'application/x-cdlink',
             '.vcg': 'application/vnd.groove-vcard', '.vcf': 'text/x-vcard', '.json': 'application/json',
             '.shf': 'application/shf+xml', '.cdbcmsg': 'application/vnd.contact.cmsg',
             '.tpt': 'application/vnd.trid.tpt', '.psb': 'application/vnd.3gpp.pic-bw-small',
             '.vxml': 'application/voicexml+xml', '.psf': 'application/x-font-linux-psf',
             '.tpl': 'application/vnd.groove-tool-template', '.htke': 'application/vnd.kenameaapp',
             '.vcx': 'application/vnd.vcx', '.teicorpus': 'application/tei+xml', '.xhtml': 'application/xhtml+xml',
             '.midi': 'audio/midi', '.tiff': 'image/tiff', '.cdmio': 'application/cdmi-object',
             '.texi': 'application/x-texinfo', '.oda': 'application/oda', '.ustar': 'application/x-ustar',
             '.ssml': 'application/ssml+xml', '.odb': 'application/vnd.oasis.opendocument.database',
             '.m3u8': 'application/x-mpegurl', '.cdmid': 'application/cdmi-domain', '.see': 'application/vnd.seemail',
             '.cdmia': 'application/cdmi-capability', '.fcs': 'application/vnd.isac.fcs',
             '.mpkg': 'application/vnd.apple.installer+xml', '.manifest': 'text/cache-manifest',
             '.odt': 'application/vnd.oasis.opendocument.text', '.3g2': 'video/3gpp2',
             '.odp': 'application/vnd.oasis.opendocument.presentation',
             '.ods': 'application/vnd.oasis.opendocument.spreadsheet', '.cdmiq': 'application/cdmi-queue',
             '.text': 'text/plain', '.qt': 'video/quicktime', '.mpm': 'application/vnd.blueice.multipass',
             '.mpc': 'application/vnd.mophun.certificate', '.mpa': 'video/mpeg', '.mpg': 'video/mpeg',
             '.mpe': 'video/mpeg', '.xdssc': 'application/dssc+xml', '.mpy': 'application/vnd.ibm.minipay',
             '.pot': 'application/vnd.ms-powerpoint', '.umj': 'application/vnd.umajin',
             '.wpd': 'application/vnd.wordperfect', '.tfi': 'application/thraud+xml',
             '.mpp': 'application/vnd.ms-project', '.tfm': 'application/x-tex-tfm', '.xspf': 'application/xspf+xml',
             '.nsf': 'application/vnd.lotus-notes', '.wmlc': 'application/vnd.wap.wmlc',
             '.ggb': 'application/vnd.geogebra.file', '.webp': 'image/webp', '.pyv': 'video/vnd.ms-playready.media.pyv',
             '.dpg': 'application/vnd.dpgraph', '.fzs': 'application/vnd.fuzzysheet',
             '.ggt': 'application/vnd.geogebra.tool', '.onetoc': 'application/onenote', '.webm': 'video/webm',
             '.kon': 'application/vnd.kde.kontour', '.karbon': 'application/vnd.kde.karbon', '.weba': 'audio/webm',
             '.pfx': 'application/x-pkcs12', '.m4a': 'audio/mp4a-latm', '.mseq': 'application/vnd.mseq',
             '.ivp': 'application/vnd.immervision-ivp', '.ivu': 'application/vnd.immervision-ivu',
             '.qti': 'image/x-quicktime', '.pfr': 'application/font-tdpfr', '.yang': 'application/yang',
             '.swa': 'application/x-director', '.gv': 'text/vnd.graphviz', '.swf': 'application/x-shockwave-flash',
             '.m4p': 'audio/mp4a-latm', '.mp3': 'audio/mpeg', '.swi': 'application/vnd.aristanetworks.swi',
             '.sldm': 'application/vnd.ms-powerpoint.slide.macroenabled.12', '.pfa': 'application/x-font-type1',
             '.pfb': 'application/x-font-type1', '.mp4': 'video/mp4', '.hvp': 'application/vnd.yamaha.hv-voice',
             '.rm': 'application/vnd.rn-realmedia', '.rl': 'application/resource-lists+xml',
             '.pkg': 'application/octet-stream', '.ra': 'audio/x-pn-realaudio', '.sbml': 'application/sbml+xml',
             '.seed': 'application/vnd.fdsn.seed', '.book': 'application/vnd.framemaker', '.cmdf': 'chemical/x-cmdf',
             '.wcm': 'application/vnd.ms-works', '.sxd': 'application/vnd.sun.xml.draw',
             '.rs': 'application/rls-services+xml', '.ppam': 'application/vnd.ms-powerpoint.addin.macroenabled.12',
             '.rq': 'application/sparql-query', '.sxg': 'application/vnd.sun.xml.writer.global',
             '.xop': 'application/xop+xml', '.skd': 'application/vnd.koan', '.xdf': 'application/xcap-diff+xml',
             '.sis': 'application/vnd.symbian.install', '.h263': 'video/h263', '.skm': 'application/vnd.koan',
             '.h261': 'video/h261', '.osfpvg': 'application/vnd.yamaha.openscoreformat.osfpvg+xml',
             '.h264': 'video/h264', '.skt': 'application/vnd.koan', '.skp': 'application/vnd.koan',
             '.ufdl': 'application/vnd.ufdl', '.for': 'text/x-fortran', '.cct': 'application/x-director',
             '.hqx': 'application/mac-binhex40', '.ttl': 'text/turtle', '.ksp': 'application/vnd.kde.kspread',
             '.sit': 'application/x-stuffit', '.doc': 'application/msword', '.srx': 'application/sparql-results+xml',
             '.wps': 'application/vnd.ms-works', '.shar': 'application/x-shar', '.psd': 'image/vnd.adobe.photoshop',
             '.ptid': 'application/vnd.pvi.ptid1', '.m3u': 'audio/x-mpegurl', '.cxt': 'application/x-director',
             '.ksh': 'text/plain', '.ccxml': 'application/ccxml+xml', '.dot': 'application/msword',
             '.cdy': 'application/vnd.cinderella', '.cdx': 'chemical/x-cdx', '.slt': 'application/vnd.epson.salt',
             '.fvt': 'video/vnd.fvt', '.vor': 'application/vnd.stardivision.writer', '.ics': 'text/calendar',
             '.vox': 'application/x-authorware-bin', '.ktr': 'application/vnd.kahootz',
             '.qps': 'application/vnd.publishare-delta-tree', '.gsf': 'application/x-font-ghostscript',
             '.icm': 'application/vnd.iccprofile', '.uoml': 'application/vnd.uoml+xml',
             '.icc': 'application/vnd.iccprofile', '.ktz': 'application/vnd.kahootz', '.uri': 'text/uri-list',
             '.cdf': 'application/x-netcdf', '.ai': 'application/postscript', '.deploy': 'application/octet-stream',
             '.mp4v': 'video/mp4', '.grv': 'application/vnd.groove-injector', '.list': 'text/plain',
             '.esf': 'application/vnd.epson.esf', '.abw': 'application/x-abiword',
             '.wspolicy': 'application/wspolicy+xml', '.mpga': 'audio/mpeg', '.epub': 'application/epub+zip',
             '.pki': 'application/pkixcmp', '.cc': 'text/x-c', '.hdf': 'application/x-hdf',
             '.davmount': 'application/davmount+xml', '.xpdl': 'application/xml', '.src': 'application/x-wais-source',
             '.zaz': 'application/vnd.zzazz.deck+xml', '.bdm': 'application/vnd.syncml.dm+wbxml',
             '.wqd': 'application/vnd.wqd', '.log': 'text/plain', '.cxx': 'text/x-c',
             '.mobipocket-ebook': 'application/octet-stream', '.box': 'application/vnd.previewsystems.box',
             '.sru': 'application/sru+xml', '.boz': 'application/x-bzip2', '.vcs': 'text/x-vcalendar',
             '.oxt': 'application/vnd.openofficeorg.extension', '.pbd': 'application/vnd.powerbuilder6',
             '.bh2': 'application/vnd.fujitsu.oasysprs', '.mpg4': 'video/mp4',
             '.pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
             '.djvu': 'image/vnd.djvu', '.pbm': 'image/x-portable-bitmap', '.es3': 'application/vnd.eszigno3+xml',
             '.qbo': 'application/vnd.intu.qbo', '.vrml': 'model/vrml', '.msty': 'application/vnd.muvee.style',
             '.dtd': 'application/xml-dtd', '.dtb': 'application/x-dtbook+xml', '.pclxl': 'application/vnd.hp-pclxl',
             '.xdp': 'application/vnd.adobe.xdp+xml', '.xdw': 'application/vnd.fujixerox.docuworks',
             '.onepkg': 'application/onenote', '.mobi': 'application/x-mobipocket-ebook',
             '.mbk': 'application/vnd.mobius.mbk', '.cdxml': 'application/vnd.chemdraw+xml', '.kar': 'audio/midi',
             '.ipk': 'application/vnd.shana.informed.package', '.org': 'application/vnd.lotus-organizer',
             '.gre': 'application/vnd.geometry-explorer', '.xslt': 'application/xslt+xml', '.aiff': 'audio/x-aiff',
             '.aifc': 'audio/x-aiff', '.xdm': 'application/vnd.syncml.dm+xml', '.gqf': 'application/vnd.grafeq',
             '.mp21': 'application/mp21', '.flx': 'text/vnd.fmi.flexstor', '.fly': 'text/vnd.fly',
             '.kne': 'application/vnd.kinar', '.edx': 'application/vnd.novadigm.edx', '.flv': 'video/x-flv',
             '.flw': 'application/vnd.kde.kivio', '.odg': 'application/vnd.oasis.opendocument.graphics',
             '.susp': 'application/vnd.sus-calendar', '.ez3': 'application/vnd.ezpix-package',
             '.ez2': 'application/vnd.ezpix-album', '.gqs': 'application/vnd.grafeq',
             '.nbp': 'application/vnd.wolfram.player', '.hbci': 'application/vnd.hbci',
             '.pkipath': 'application/pkix-pkipath', '.uvvd': 'application/vnd.dece.data',
             '.odc': 'application/vnd.oasis.opendocument.chart', '.uvvf': 'application/vnd.dece.data',
             '.sema': 'application/vnd.sema', '.uvvh': 'video/vnd.dece.hd', '.uvvi': 'image/vnd.dece.graphic',
             '.ppt': 'application/vnd.ms-powerpoint', '.pps': 'application/vnd.ms-powerpoint',
             '.uvvp': 'video/vnd.dece.pd', '.odm': 'application/vnd.oasis.opendocument.text-master',
             '.ppm': 'image/x-portable-pixmap', '.lzh': 'application/octet-stream', '.uvvu': 'video/vnd.uvvu.mp4',
             '.uvvv': 'video/vnd.dece.video', '.xvm': 'application/xv+xml', '.latex': 'application/x-latex',
             '.semd': 'application/vnd.semd', '.ppd': 'application/vnd.cups-ppd', '.cgm': 'image/cgm',
             '.ppa': 'application/vnd.ms-powerpoint', '.fpx': 'image/vnd.fpx', '.igl': 'application/vnd.igloader',
             '.mbox': 'application/mbox', '.lwp': 'application/vnd.lotus-wordpro',
             '.osf': 'application/vnd.yamaha.openscoreformat', '.gdl': 'model/vnd.gdl',
             '.odi': 'application/vnd.oasis.opendocument.image', '.kwt': 'application/vnd.kde.kword',
             '.dcr': 'application/x-director', '.oti': 'application/vnd.oasis.opendocument.image-template',
             '.mp2a': 'audio/mpeg', '.cdmic': 'application/cdmi-container', '.igx': 'application/vnd.micrografx.igx',
             '.kwd': 'application/vnd.kde.kword', '.igs': 'model/iges', '.apr': 'application/vnd.lotus-approach',
             '.crt': 'application/x-x509-ca-cert', '.n-gage': 'application/vnd.nokia.n-gage.symbian.install',
             '.tcl': 'application/x-tcl', '.ogx': 'application/ogg', '.flo': 'application/vnd.micrografx.flo',
             '.jam': 'application/vnd.jam', '.rlc': 'image/vnd.fujixerox.edmics-rlc',
             '.rld': 'application/resource-lists-diff+xml', '.svgz': 'image/svg+xml', '.bz2': 'application/x-bzip2',
             '.xar': 'application/vnd.xara', '.jar': 'application/java-archive', '.cpt': 'application/mac-compactpro',
             '.ogg': 'audio/ogg', '.dts': 'audio/vnd.dts', '.oga': 'audio/ogg', '.afp': 'application/vnd.ibm.modcap',
             '.f90': 'text/x-fortran', '.application': 'application/x-ms-application', '.mxf': 'application/mxf',
             '.rgb': 'image/x-rgb', '.mxl': 'application/vnd.recordare.musicxml',
             '.mxs': 'application/vnd.triscape.mxs', '.gram': 'application/srgs', '.me': 'text/troff',
             '.mb': 'application/mathematica', '.mxu': 'video/vnd.mpegurl', '.ma': 'application/mathematica',
             '.qam': 'application/vnd.epson.quickanime', '.afm': 'application/x-font-type1',
             '.thmx': 'application/vnd.ms-officetheme', '.mseed': 'application/vnd.fdsn.mseed', '.fli': 'video/x-fli',
             '.asx': 'video/x-ms-asf', '.pcurl': 'application/vnd.curl.pcurl',
             '.ser': 'application/java-serialized-object', '.portpkg': 'application/vnd.macports.portpkg',
             '.vsf': 'application/vnd.vsf', '.res': 'application/x-dtbresource+xml',
             '.pgp': 'application/pgp-encrypted', '.svc': 'application/vnd.dvb.service',
             '.nlu': 'application/vnd.neurolanguage.nlu', '.pgm': 'image/x-portable-graymap', '.xyz': 'chemical/x-xyz',
             '.svg': 'image/svg+xml', '.svd': 'application/vnd.svd', '.atom': 'application/atom+xml',
             '.qfx': 'application/vnd.intu.qfx', '.roff': 'text/troff', '.unityweb': 'application/vnd.unity',
             '.123': 'application/vnd.lotus-1-2-3', '.xlsm': 'application/vnd.ms-excel.sheet.macroenabled.12',
             '.uvd': 'application/vnd.dece.data', '.xbap': 'application/x-ms-xbap', '.uvf': 'application/vnd.dece.data',
             '.uva': 'audio/vnd.dece.audio', '.uvm': 'video/vnd.dece.mobile', '.scurl': 'text/vnd.curl.scurl',
             '.uvi': 'image/vnd.dece.graphic', '.uvh': 'video/vnd.dece.hd', '.eol': 'audio/vnd.digital-winds',
             '.frame': 'application/vnd.framemaker', '.uvu': 'video/vnd.uvvu.mp4',
             '.uvt': 'application/vnd.dece.ttml+xml', '.lostxml': 'application/lost+xml', '.spx': 'audio/ogg',
             '.man': 'text/troff', '.mac': 'image/x-macpaint', '.uvx': 'application/vnd.dece.unspecified',
             '.mag': 'application/vnd.ecowin.chart', '.pya': 'audio/vnd.ms-playready.media.pya',
             '.pyc': 'application/x-python-code', '.atx': 'application/vnd.antix.game-component',
             '.dataless': 'application/vnd.fdsn.seed', '.apk': 'application/vnd.android.package-archive',
             '.mny': 'application/x-msmoney', '.pyo': 'application/x-python-code',
             '.ghf': 'application/vnd.groove-help', '.cpio': 'application/x-cpio',
             '.setreg': 'application/set-registration-initiation', '.atc': 'application/vnd.acucorp',
             '.cml': 'chemical/x-cml', '.ecelp4800': 'audio/vnd.nuera.ecelp4800', '.xpi': 'application/x-xpinstall',
             '.jnlp': 'application/x-java-jnlp-file', '.uvva': 'audio/vnd.dece.audio',
             '.dll': 'application/x-msdownload', '.rif': 'application/reginfo+xml', '.uvvm': 'video/vnd.dece.mobile',
             '.scq': 'application/scvp-cv-request', '.scs': 'application/scvp-cv-response',
             '.scm': 'application/vnd.lotus-screencam', '.xfdl': 'application/vnd.xfdl', '.rip': 'audio/vnd.rip',
             '.scd': 'application/x-msschedule', '.wdb': 'application/vnd.ms-works',
             '.xfdf': 'application/vnd.adobe.xfdf', '.xwd': 'image/x-xwindowdump',
             '.xlam': 'application/vnd.ms-excel.addin.macroenabled.12', '.aab': 'application/x-authorware-bin',
             '.aac': 'audio/x-aac', '.sda': 'application/vnd.stardivision.draw', '.aam': 'application/x-authorware-map',
             '.sdc': 'application/vnd.stardivision.calc', '.g3': 'image/g3fax',
             '.sdd': 'application/vnd.stardivision.impress', '.js': 'application/javascript',
             '.aas': 'application/x-authorware-seg', '.sdp': 'application/sdp',
             '.sdw': 'application/vnd.stardivision.writer', '.plb': 'application/vnd.3gpp.pic-bw-large',
             '.plc': 'application/vnd.mobius.plc', '.plf': 'application/vnd.pocketlearn', '.3dml': 'text/vnd.in3d.3dml',
             '.lha': 'application/octet-stream', '.edm': 'application/vnd.novadigm.edm',
             '.wtb': 'application/vnd.webturbo', '.msf': 'application/vnd.epson.msf', '.pls': 'application/pls+xml',
             '.knp': 'application/vnd.kinar', '.msl': 'application/vnd.mobius.msl',
             '.mpt': 'application/vnd.ms-project', '.qxb': 'application/vnd.quark.quarkxpress', '.msh': 'model/mesh',
             '.msi': 'application/x-msdownload', '.oprc': 'application/vnd.palm', '.pcf': 'application/x-font-pcf',
             '.xla': 'application/vnd.ms-excel', '.potm': 'application/vnd.ms-powerpoint.template.macroenabled.12',
             '.7z': 'application/x-7z-compressed', '.pcl': 'application/vnd.hp-pcl', '.jpe': 'image/jpeg',
             '.jpg': 'image/jpeg', '.mdb': 'application/x-msaccess', '.cmp': 'application/vnd.yellowriver-custom-menu',
             '.potx': 'application/vnd.openxmlformats-officedocument.presentationml.template', '.jpm': 'video/jpm',
             '.pcx': 'image/x-pcx', '.zir': 'application/vnd.zul', '.mdi': 'image/vnd.ms-modi',
             '.zip': 'application/zip', '.vss': 'application/vnd.visio', '.adp': 'audio/adpcm',
             '.clkk': 'application/vnd.crick.clicker.keyboard',
             '.ppsx': 'application/vnd.openxmlformats-officedocument.presentationml.slideshow',
             '.vsw': 'application/vnd.visio', '.class': 'application/java-vm', '.torrent': 'application/x-bittorrent',
             '.mmf': 'application/vnd.smaf', '.mj2': 'video/mj2', '.m3a': 'audio/mpeg',
             '.xer': 'application/patch-ops-error+xml', '.asc': 'application/pgp-signature',
             '.qxl': 'application/vnd.quark.quarkxpress', '.fh4': 'image/x-freehand', '.vsd': 'application/vnd.visio',
             '.clkx': 'application/vnd.crick.clicker', '.clkw': 'application/vnd.crick.clicker.wordbank',
             '.xo': 'application/vnd.olpc-sugar', '.clkt': 'application/vnd.crick.clicker.template',
             '.aso': 'application/vnd.accpac.simply.aso', '.mmr': 'image/vnd.fujixerox.edmics-mmr',
             '.pwz': 'application/vnd.ms-powerpoint', '.sh': 'application/x-sh',
             '.sm': 'application/vnd.stepmania.stepchart', '.hvd': 'application/vnd.yamaha.hv-dic',
             '.so': 'application/octet-stream', '.wbs': 'application/vnd.criticaltools.wbs+xml',
             '.xlc': 'application/vnd.ms-excel', '.tcap': 'application/vnd.3gpp2.tcap',
             '.sfd-hdstx': 'application/vnd.hydrostatix.sof-data', '.cif': 'chemical/x-cif',
             '.xls': 'application/vnd.ms-excel', '.cii': 'application/vnd.anser-web-certificate-issue-initiation',
             '.st': 'application/vnd.sailingtracker.track', '.xlw': 'application/vnd.ms-excel',
             '.cil': 'application/vnd.ms-artgalry', '.spp': 'application/scvp-vp-response',
             '.wmls': 'text/vnd.wap.wmlscript', '.wks': 'application/vnd.ms-works', '.wmv': 'video/x-ms-wmv',
             '.gtw': 'model/vnd.gtw', '.uvvg': 'image/vnd.dece.graphic', '.fh7': 'image/x-freehand',
             '.gtm': 'application/vnd.groove-tool-message', '.trm': 'application/x-msterminal',
             '.ppsm': 'application/vnd.ms-powerpoint.slideshow.macroenabled.12', '.urls': 'text/uri-list',
             '.woff': 'application/x-font-woff', '.html': 'text/html', '.kia': 'application/vnd.kidspiration',
             '.odf': 'application/vnd.oasis.opendocument.formula', '.tra': 'application/vnd.trueapp',
             '.fsc': 'application/vnd.fsc.weblaunch', '.ifm': 'application/vnd.shana.informed.formdata',
             '.mime': 'message/rfc822', '.sgl': 'application/vnd.stardivision.writer-global', '.sgm': 'text/sgml',
             '.musicxml': 'application/vnd.recordare.musicxml+xml',
             '.lbd': 'application/vnd.llamagraphics.life-balance.desktop',
             '.lbe': 'application/vnd.llamagraphics.life-balance.exchange+xml', '.maker': 'application/vnd.framemaker',
             '.fst': 'image/vnd.fst', '.wm': 'video/x-ms-wm', '.tif': 'image/tiff', '.n3': 'text/n3',
             '.otp': 'application/vnd.oasis.opendocument.presentation-template', '.dv': 'video/x-dv',
             '.qwt': 'application/vnd.quark.quarkxpress', '.f77': 'text/x-fortran', '.pic': 'image/x-pict',
             '.mvb': 'application/x-msmediaview', '.les': 'application/vnd.hhe.lesson-player',
             '.uvvs': 'video/vnd.dece.sd', '.rms': 'application/vnd.jcp.javame.midlet-rms',
             '.uvvt': 'application/vnd.dece.ttml+xml', '.mets': 'application/mets+xml',
             '.dir': 'application/x-director', '.acutc': 'application/vnd.acucorp', '.rdf': 'application/rdf+xml',
             '.xsl': 'application/xml', '.sxw': 'application/vnd.sun.xml.writer', '.spot': 'text/vnd.in3d.spot',
             '.nc': 'application/x-netcdf', '.nb': 'application/mathematica', '.sxm': 'application/vnd.sun.xml.math',
             '.uvvx': 'application/vnd.dece.unspecified', '.sxi': 'application/vnd.sun.xml.impress',
             '.aep': 'application/vnd.audiograph', '.bin': 'application/octet-stream',
             '.rdz': 'application/vnd.data-vision.rdz', '.uris': 'text/uri-list',
             '.sxc': 'application/vnd.sun.xml.calc', '.h': 'text/x-c', '.ac': 'application/pkix-attr-cert',
             '.pptm': 'application/vnd.ms-powerpoint.presentation.macroenabled.12',
             '.itp': 'application/vnd.shana.informed.formtemplate',
             '.xltm': 'application/vnd.ms-excel.template.macroenabled.12', '.o': 'application/octet-stream',
             '.a': 'application/octet-stream', '.c': 'text/x-c', '.rcprofile': 'application/vnd.ipunplugged.rcprofile',
             '.udeb': 'application/x-debian-package', '.f': 'text/x-fortran',
             '.fe_launch': 'application/vnd.denovo.fcselayout-link', '.vtu': 'model/vnd.vtu',
             '.xltx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
             '.aw': 'application/applixware', '.au': 'audio/basic', '.p': 'text/x-pascal', '.s': 'text/x-asm',
             '.t': 'text/troff', '.wpl': 'application/vnd.ms-wpl', '.cpp': 'text/x-c', '.xenc': 'application/xenc+xml',
             '.btif': 'image/prs.btif', '.gnumeric': 'application/x-gnumeric', '.mods': 'application/mods+xml',
             '.pct': 'image/x-pict', '.docm': 'application/vnd.ms-word.document.macroenabled.12',
             '.nnw': 'application/vnd.noblenet-web', '.xvml': 'application/xv+xml',
             '.x3d': 'application/vnd.hzn-3d-crossword', '.xap': 'application/x-silverlight-app',
             '.igm': 'application/vnd.insors.igm',
             '.sldx': 'application/vnd.openxmlformats-officedocument.presentationml.slide',
             '.ttc': 'application/x-font-ttf',
             '.docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
             '.zmm': 'application/vnd.handheld-entertainment+xml', '.sv4crc': 'application/x-sv4crc',
             '.cmc': 'application/vnd.cosmocaller', '.pfm': 'application/x-font-type1',
             '.u32': 'application/x-authorware-bin', '.m4u': 'video/vnd.mpegurl',
             '.pvb': 'application/vnd.3gpp.pic-bw-var', '.jpeg': 'image/jpeg',
             '.mgz': 'application/vnd.proteus.magazine', '.mid': 'audio/midi', '.mif': 'application/vnd.mif',
             '.chat': 'application/x-chat', '.cmx': 'image/x-cmx', '.kml': 'application/vnd.google-earth.kml+xml',
             '.dump': 'application/octet-stream', '.gph': 'application/vnd.flographit', '.xht': 'application/xhtml+xml',
             '.dif': 'video/x-dv', '.grxml': 'application/srgs+xml', '.dic': 'text/x-c',
             '.ncx': 'application/x-dtbncx+xml', '.wg': 'application/vnd.pmi.widget',
             '.kmz': 'application/vnd.google-earth.kmz', '.mp2': 'audio/mpeg', '.x32': 'application/x-authorware-bin',
             '.teacher': 'application/vnd.smart.teacher', '.dis': 'application/vnd.mobius.dis',
             '.vst': 'application/vnd.visio', '.curl': 'text/vnd.curl', '.setpay': 'application/set-payment-initiation',
             '.cod': 'application/vnd.rim.cod', '.w3d': 'application/x-director', '.snf': 'application/x-font-snf',
             '.snd': 'audio/basic', '.mmd': 'application/vnd.chipnuts.karaoke-mmd', '.pict': 'image/pict',
             '.mjp2': 'video/mj2', '.lasxml': 'application/vnd.las.las+xml', '.ftc': 'application/vnd.fluxtime.clip',
             '.pqa': 'application/vnd.palm', '.java': 'text/x-java-source',
             '.fti': 'application/vnd.anser-web-funds-transfer-initiation', '.mads': 'application/mads+xml',

             '.less':'text/less', '.sass':'text/sass', '.scss': 'text/scss',

             '.gz': 'application/gzip',
             '.md': 'application/markdown',
             '.markdown': 'application/markdown',
             '.mk': 'application/markdown',
             '.config': 'application/config',

             '.coffee': 'text/x-coffeescript',


}

def guess_type(name, default_type=''):
    if not name:
        return default_type
    ext = os.path.splitext(name)[1].lower()
    if '?' in ext:
        ext = ext.split('?')[0]
    return MIMETYPES.get(ext) or default_type


#coding: utf8
from __future__ import absolute_import
import os


def get_env(key):
    lower_key = key.lower()
    v = os.environ.get(key) or os.environ.get(lower_key)
    if v:
        return v
    filenames  = [key, '%s.txt'%key]
    if lower_key not in filenames:
        filenames += [lower_key, '%s.txt'%lower_key]
    for filename in filenames:
        filepath = os.path.join('/tmp/env', filename)
        if os.path.isfile(filepath):
            with open(filepath, 'rb') as f:
                raw_content = f.read()
            v = raw_content.strip()
            if v:
                return v







# coding: utf8
import time
from flask import g, request


def set_cookies(response):
    # the value must dumps by singer
    # 在website.core中调用，after_request，可以处理安全性cookie的读写删除
    cookies_to_set = getattr(g, 'cookies', {})
    for k, v in cookies_to_set.items():
        # 用expires, 而非max_age，这样可以重复set_cookie, 即使nano在proxy的情况下
        response.set_cookie(k, v, expires=time.time()+10*60)

    # delete the cookies
    cookies_to_delete = getattr(g, 'cookies_to_delete', [])
    for key in cookies_to_delete:
        if key not in cookies_to_set and key in request.cookies:
            response.delete_cookie(key)


def default_response_cookie_handler(response):
    set_cookies(response)

    if getattr(g, 'response_content_type', None):
        response.content_type = g.response_content_type

    return response



#coding: utf8
from __future__ import absolute_import
import os
import re
from farbox_lite.utils import same_slash, smart_unicode



def get_index_doc_filepath(folder_path):
    md_index_filenames = ['index.md', 'index.txt', 'index.mk', 'index.markdown']
    for filename in md_index_filenames:
        filepath = os.path.join(folder_path, filename)
        if os.path.isfile(filepath):
            return filepath
    return None # at last



def is_real(path):
    # 主要是判断是否真实的文档，还是软链，或者软链下的目录内
    path = same_slash(path)
    if not os.path.exists(path):
        return False
    parts = path.split('/')
    for i in range(len(parts)):
        if i:
            _path = '/'.join(parts[:-i])
        else:
            _path = path
        if os.path.islink(_path):
            return False
    return True


def is_a_hidden_path(path):
    path = same_slash(path)
    if re.search('(^|/)(\.|~$)', path):
        return True
    elif re.search(r'~\.[^.]+$', path):
        return True
    elif path.endswith('~'):
        return True
    else:
        return False


def is_a_empty_folder(folder_path, include_empty_sub_folder=True):
    if not os.path.isdir(folder_path):
        return False
    for sub_name in os.listdir(folder_path):
        file_path = os.path.join(folder_path, sub_name)
        if not is_a_hidden_path(file_path) and os.path.exists(file_path):
            if os.path.isdir(file_path) and is_a_empty_folder(file_path) and include_empty_sub_folder:
                # 如果包含的目录也是空目录，不处理， continue
                continue
            return False
    # at last
    return True



def get_file_list(root_path, split=False, filter_func=None, parent_filter_func=None):
    root_path = same_slash(root_path)
    file_paths = []
    just_files = []
    just_folders = []
    if not os.path.isdir(root_path): # 根目录不存在，不处理
        pass
    else:
        for parent, folders, files in os.walk(root_path):
            if is_a_hidden_path(parent):
                continue
            elif not is_real(parent): # link类型的不处理
                continue

            if parent_filter_func:
                if not parent_filter_func(parent): # parent 不需要处理的, 直接 ignore
                    continue

            for filename in files:
                file_path = os.path.join(parent, filename)
                if not is_a_hidden_path(file_path) and is_real(file_path):
                    #relative_path = file_path.replace(root_path, '').strip('/')
                    file_paths.append(file_path)
                    just_files.append(file_path)

            for filename in folders:
                file_path = os.path.join(parent, filename)
                if not is_a_hidden_path(file_path) and is_real(file_path):
                    #relative_path = file_path.replace(root_path, '').strip('/')
                    file_paths.append(file_path)
                    just_folders.append(file_path)
    if filter_func:
        file_paths = [path for path in file_paths if filter_func(path)]
        just_folders = [path for path in just_folders if filter_func(path)]

    if not split:
        return file_paths
    else:
        return just_folders, just_files


def get_sort_and_name_from_filepath(filepath):
    filename = os.path.split(filepath)[-1]
    if os.path.isdir(filepath):
        name = filename
    else:
        name = os.path.splitext(filename)[0]
    name = name.strip()
    sort = -1
    if re.match(r'\d+ ', name) or re.match(r'\d+\.\d+ ', name):
        sort, name = name.split(' ', 1)
        sort = sort.strip()
        if '.' in sort:
            try: sort = float(sort)
            except: pass
        else:
            try: sort = int(sort)
            except: pass
    return sort, name






#coding: utf8
from pyjade.utils import process
from pyjade import Compiler as _Compiler
from pyjade.parser import Parser
from pyjade.lexer import Lexer
import re, os
from jinja2 import Template
from farbox_lite.utils import smart_unicode


def re_exclude_split(text, split_str, exclude_rule='\(.*?\)', maxsplit=0):
    # 某段 text 按照split_str分割成 list，但是某个正则规则内（比如括号内的）的split_str 则不能计算在内
    # split_str = ':'
    attrs_parts = []
    for i in re.finditer(exclude_rule, text):
        attrs_parts += range(i.start(), i.end())

    split_parts = [] # 括号内的 : 是不能用于分割的
    split_count = 0
    for split_found in re.finditer(split_str, text):
        if split_found.start() not in attrs_parts:
            split_parts.append(split_found.start())
            split_count += 1
            if maxsplit and split_count >= split_count:
                break

    parts = []
    if split_parts:
        indexes_parts = zip([0]+split_parts, split_parts+[len(text)])
        for i,j in indexes_parts:
            part = text[i:j]
            sub_pattern = split_str
            if not sub_pattern.startswith('^'):
                sub_pattern = '^'+sub_pattern
            part = re.sub(sub_pattern, '', part) # 清除参与分割的字符串
            parts.append(part)

    return parts

class JadCompilerError(Exception):
    pass

lexer_init = Lexer.__init__

def _lexer_init(self, *args, **kwargs):
    lexer_init(self, *args, **kwargs)
    self.RE_ATTR_INTERPOLATE = re.compile(r'#\{([^{}]+)\}')

    # 避免有';'的时候，被隔断，例如： row_style = 'background:url(%s);'%row.bg.value
    self.RE_ASSIGNMENT = re.compile(r'^(-\s+var\s+)?(\w+) += *([^\n]+)( *;? *)')

Lexer.__init__ = _lexer_init

parseExpr = Parser.parseExpr

def _parser_except(self, type):
    t = self.peek().type
    if t == type:
        return self.advance()
    else:
        raw_content = re.sub(r'\r\n|\r', '\n', self.input)
        compiled_content = self.input[:raw_content.rfind(self.lexer.input)]
        line_number = re.sub(r'\s+$', '', compiled_content).count('\n') + 1
        raise JadCompilerError('expected "%s" but got "%s" on line %d' %  (type, t, line_number))

def _parseExpr(self):
    t = self.peek().type
    try:
        return parseExpr(self)
    except Exception, e:
        if not isinstance(e, JadCompilerError):
            raw_content = re.sub(r'\r\n|\r', '\n', self.input)
            compiled_content = self.input[:raw_content.rfind(self.lexer.input)]
            line_number = re.sub(r'\s+$', '', compiled_content).count('\n') + 1
            raise JadCompilerError('unexpected token "%s" on line %d'  %  (t, line_number))
        else:
            raise e


Parser.parseExpr = _parseExpr
Parser.expect = _parser_except


ONE_LINE_FUNCTIONS = ['load', 'set_content_type', 'redirect', 'a_with_selected', 'set_per_page', 'add_doc_actions', 'make_site_live']
DEFAULT_NAMES = ['request','response', 'doc','post', 'posts', 'site', 'sites', 'files', 'tags', 'tag', 'folder', 'folders',
                    'category', 'images', 'albums', 'album', 'next_one', 'pre_one', 'paginator', 'pager']


block_or_re_pattern = 'page|scroll|font|browser|pure|footer|create_dom|modal|tab' # block 语法

class Compiler(_Compiler):
    def __init__(self, node, **options):
        _Compiler.__init__(self, node, **options)
        self.current_line = 0

    def visit(self, node, *args, **kwargs):
        result = _Compiler.visit(self, node, *args,**kwargs)
        if hasattr(node, 'line') and node.line != self.current_line:
            self.current_line = node.line
            self.buffer('\n')
        return result

    def buffer(self,strs):
        if strs.startswith('{%'):
            strs=  '\n%s\n' % strs
        _Compiler.buffer(self, strs)

    def visitText(self,text):
        text = ''.join(text.nodes)
        text = self.interpolate(text)
        self.buffer(text) # 不用补充 \n 了，在visit 函数中会自动处理

    def visitCodeBlock(self,block):
      self.buffer('{%% block %s %%}'%block.name)
      if block.mode=='append': self.buffer('%ssuper()%s' % (self.variable_start_string, self.variable_end_string))
      self.visitBlock(block)
      if block.mode=='prepend': self.buffer('%ssuper()%s' % (self.variable_start_string, self.variable_end_string))
      self.buffer('{% endblock %}')

    def visitTag(self, tag):
        name = tag.name
        if name in DEFAULT_NAMES or (name.startswith('_') and len(name)>1):
            # 相当于按照我们自己的声明，调用 backend 的某个属性或者函数
            # 这些tag的，可以处理不带函数的变量
            # not tag.attrs 就是单纯一个变量的调用....
            #  {'name': 'class', 'static': True, 'val': u'"bcd tbc"'}]
            var_name_list = [name.lstrip('_')]
            if not tag.attrs: # 单纯变量调用，没有子属性，也不可能是函数调用
                self.buffer('{{ %s }}' % '.'.join(var_name_list))
            else: # tag.attrs 是有序的 list
                attrs = []
                class_attr = None
                for attr in tag.attrs:
                    if attr.get('name') == 'class': # 子属性的调用实际上是
                        class_attr = attr
                    else:
                        attrs.append(attr)
                if class_attr:
                    var_name_list += class_attr.get('val', '').strip('"\'').split(' ') # 原始会被视为 classes
                var_real_name = '.'.join(var_name_list)

                # 可能是赋值
                text_notes = getattr(getattr(tag, 'text', None), 'nodes', None)
                if text_notes and len(text_notes) == 1:
                    text = text_notes[0].strip()
                    if len(text) >=2 and text[0]=='=' and text[1]!='=':
                        # 实际上赋值行为
                        if '.' not in var_real_name:
                            return self.buffer('{%% set %s = %s %%}' % (var_real_name,text[1:]))
                        else: # 子属性的赋值是 jinja 不支持的，所以调用一个公用函数 set_property 这非常重要！
                            parent_obj_name = '.'.join(var_name_list[:-1])
                            property_name = var_name_list[-1]
                            return self.buffer('{{ set_property(%s, "%s", %s) }}' % (parent_obj_name, property_name, text[1:]))

                self.buffer('{{ %s }}' % var_real_name)

        elif name in ['from'] and hasattr(tag, 'text') and hasattr(tag.text, 'nodes') and len(tag.text.nodes)==1:
            # from xxx import 类似的，
            sub_text = tag.text.nodes[0].strip()
            parts = re.split(' +', sub_text, 1)
            if len(parts)==2 and '"' not in parts[1] and "'" not in parts[1]:
                p1, p2 = parts
                current_line = '{%% from "%s" %s %%}' %(p1.strip('\'"'), p2)
                self.buffer(current_line)
            else:
                return _Compiler.visitTag(self, tag)
        else:
            return _Compiler.visitTag(self, tag)

    def visitMixin(self,mixin):
        if not mixin.call:
            self.buffer('{%% macro %s(%s) %%}'%(mixin.name,mixin.args))
            self.visitBlock(mixin.block)
            self.buffer('\n{% endmacro %}\n')
        else:
            if not mixin.block:
                self.buffer('{{ %s(%s) }}' % (mixin.name, mixin.args))
            else:
                self.buffer('{%% call %s(%s) %%}'%(mixin.name,mixin.args))
                self.visitBlock(mixin.block)
                self.buffer('\n{% endcall %}\n')


    def visitAssignment(self,assignment):
        self.buffer('{%% set %s = %s %%}'%(assignment.name, assignment.val))

    def interpolate(self,text):
        return self._interpolate(text,lambda x:'{{ %s }}' % x)

    def visitVar(self, var, escape=False):
        var = self.var_processor(var)
        return '{{ %s }}' % var

    def visitEach(self,each):
        self.buf.append("{%% for %s in %s %%}"%(','.join(each.keys), each.obj))
        self.visit(each.block)
        self.buf.append('{% endfor %}')


    def visitDynamicAttributes(self,attrs):
        buf = ''
        for attr in attrs:
            buf += ' %s="{{%s}}" ' % (attr['name'], attr['val'])
        self.buf.append(buf)


#FEED_HTML = '\n<link rel="alternate" type="application/rss+xml" title="atom 1.0" href="/feed">\n'
CONTENT_TYPE_HTML = '\n<meta http-equiv="content-type" content="text/html; charset=utf-8">\n'




def beautiful_jade(source):
    # 先处理跨行的，合并为单行
    source = re.sub(r'\\[\t \r]*\n\s*', '', source)

    # 跨行的赋值处理
    multi_lines_vars = re.findall('{%[\t \r\n].*?[\t \r\n]%}', source, re.S|re.M)
    for multi_lines_var in multi_lines_vars:
        if '\n' in multi_lines_var:
            source = source.replace(multi_lines_var, re.sub(r'[\r\n]', '', multi_lines_var), 1)

    source = source.replace('\r\n', '\n')
    source = re.sub(r'\t', '    ', source) # 替换掉 tab 键为空格
    raw_lines = re.split(r'\n|\r', source)
    lines = []
    raw_pre_space_count = 0 # 原始文本的上行空格数
    space_count_added_points = {} # 因为 : 分行的原因，在某个缩进位置产生的行首叠加
    # {4:8} 表示在第一个4格缩进的时候，增加了8个 offset，在下载4格缩进或者更小的情况下，清空这个 point
    # {4:4,  8:4} 意味着实际在8+4的位置上，是 4+4的偏移

    in_special_block = False # like style or script
    special_block_space = 0 # 开始进入的节点


    for line_i, line in enumerate(raw_lines):
        # 去除了括号内的内容的 line，这样比较容易判断 dom 本身的结构
        line_without_attrs = re.sub(r'\(.*?\)', '', line).strip()
        line_strip = line.strip()

        if not line_strip: # 空行，不做处理, 不然会影响缩进的计算
            lines.append(line)
            continue

        # 获得行首空格的数目, 先集成之前行的
        space_count = raw_pre_space_count

        space_c = re.search(r'^ +', line)
        if space_c:
            raw_space_count = space_count = len(space_c.group())
        else:
            raw_space_count = 0


        # 一些特殊代码块进入的状态判断
        if raw_space_count > raw_pre_space_count:
            # 自级
            if line_i:
                pre_line = raw_lines[line_i-1]
                if re.match(r'\s*(style|script)(\(|\s*$)', pre_line) and not in_special_block:
                    # 一些特殊节点下面的，不再做split_colon 的处理
                    in_special_block = True
                    special_block_space = raw_pre_space_count
        else:
            if raw_space_count <= special_block_space:
                in_special_block = False
                special_block_space = 0


        #print line, in_special_block


        points = space_count_added_points.items()
        offset = 0
        for space_count_added_at, space_count_added_offset in points:
            if raw_space_count <= space_count_added_at:
                space_count_added_points.pop(space_count_added_at, None)
            else:
                offset += space_count_added_offset

        if raw_pre_space_count >= raw_space_count:
            #上行的缩进多余当前行，当前 level 相当于另起一行了，使用raw_space_count
            space_count = raw_space_count

        # 偏差性移动
        space_count += offset

        split_colon = False # 将 : 替换被必要的换行
        if re.search(r'(^| )(if|for|else|elif) .*?: ', line) or re.search(r': (if|for|else|elif) ', line):
            split_colon = True
        elif re.match(r'([\.#]?[a-z])', line_without_attrs, re.I) and ":" in line_without_attrs \
                and re.search(r'[a-z0-9](= [\'"].*?[\'"])?$', line_without_attrs, re.I):
            # 该行是 dom 结构的父子关系的声明，不是a= v or a = v 之类的, 也不是 +, _开头的函数调用和变量调用
            # 当然 post:xxxx 这种形式不要出现，作为变量的 post，在此时就是语法冲突的
            # li: a= post.title 这个是可以的
            split_colon = True
        elif re.match(r'\+(%s)(\(.*?\))?\s*:'%block_or_re_pattern, line_strip): # 代码块语法的，也允许冒号进行分割
            split_colon = True
        if split_colon: # 不分割的特殊情况
            if re.search(r'style\s*=', line): # style 申明 in one line
                split_colon = False
            elif in_special_block:
                split_colon = False

        if split_colon:
            if re.match(r'if |else\s*:|elif |else if ', line_strip): # 逻辑判断的，必须分割
                pass
            elif ('+' in line and not line_strip.startswith('+')) or ' = ' in line:
                # 可能是一些变量里比如 style，带来 : , 为了避免麻烦
                split_colon = False

        if split_colon:
            # post_url = 'http://' + request.host + post.url.escaped  这个就不算: 能分割的了
            dom_parts = re_exclude_split(line_strip, ':', exclude_rule='[\(\[\"\'].*?[\]\)\"\']')
            dom_parts = [part.lstrip() for part in dom_parts]

            if dom_parts:
                # 记录偏移点，实际当前位置是不会增加偏移的，只会继承之前的偏移
                space_count_added_points[raw_space_count] = 4 * (len(dom_parts)-1)
                for dom_part in dom_parts:
                    lines.append(' '*space_count+dom_part)
                    space_count += 4
            else: # 实际上没有缩进
                lines.append(' '*space_count+line_strip)
        else:
            # 多变量赋值
            if ', ' in line_strip and ' = ' in line_strip and line_strip.count(', ')+1 == line_strip.count(' = '):
                sub_lines = line_strip.split(', ')
                for sub_line in sub_lines:
                    sub_line = sub_line.strip()
                    if not sub_line: continue
                    lines.append(' '*space_count+sub_line)
            else: #rest 获得新的 space_count 组成新的 line
                lines.append(' '*space_count+line_strip)


        # at last
        raw_pre_space_count = raw_space_count

    new_lines = []
    for line in lines:
        # 全局通用的变量空间
        global_types = 'posts|images|request|response'
        # todo + category & etc

        line_strip = line.strip()
        if line_strip.startswith('///'):
            # 完全注释，都不会出现在 html 的代码中
            continue
        elif line_strip.startswith('+'):
            # 函数的调用，不管是全局变量，都改成 jinja 先
            # 比如 +category.set('text', category['title']) 就会不认...
            # 变量的调用也可以用+的形式来
            if line_strip[1:] in ['caller']:
                new_line = line
            elif not re.match('\+\w+\(.*?\)$', line_strip) and not re.match('\+(%s)'%block_or_re_pattern, line_strip):
                # 非函数调用需要属性的（call），非 pure 等 block 代码语法，自动处理为变量先
                new_line = re.sub(r'(\s*)(\+)(\w+.*)', '\g<1>{{ \g<3> }}', line)
            else:
                #  +xxx()  这种纯 call 的，直接返回，可能是 mixin 的内部
                new_line = line
        else:
            # 函数的调用，这里会以)结尾; 预先转为 jinja 模式，不然在 jade 里没有办法解析

            # by default
            new_line = line
            done = False

            # 可能是 title= xxx 这种前后端结合的模式， 比如下面的例子，虽然又 posts.tag..join('+')， 但不能把它单独{{}} 起来
            # title= posts.keywords or post.title or posts.tags.join('+') or posts.category.title or site.title
            if line.count('= ') == 1 and ' = ' not in line:
                front_part, end_part = new_line.split('= ')
                front_part_strip = front_part.strip()
                if (re.match(r'\w+$', front_part_strip) or re.match(r'\w+\(.*?\)$', front_part_strip)) \
                        and not end_part.strip().startswith('{{'):
                    new_line = front_part + '{{ %s }}' % end_part
                    done = True

            if not done and not re.match(r'\s*\w+(\.\w+)? = ', line) and not re.search('%s\.\w+\s*\}\}' % global_types, line):
                # 替换全局变量名
                # 但不能是   xx = xx 这种赋值类型的
                # 也不能是 a(target="_blank", href="http://{{request.domain}}") 类似已经处理了的
                new_line = re.sub(r'(\s*)(_)?(%s)(\..*?\)\s*)' % global_types, '\g<1>{{ \g<3>\g<4> }}', line)



        new_lines.append(new_line)

    new_source = '\n'.join(new_lines)


    # 语法 block 的特别处理
    new_source = re.sub(r'(^ *|\n *)\+(%s)( *\n)' % block_or_re_pattern, '\g<1>+\g<2>()\g<3>', new_source)

    return new_source





def exchange(source):
    source = smart_unicode(source)
    source = source.strip().replace(u'\ufeff', '') #去除头尾

    source = re.sub(r'\\\r?\n', '', source) # 多行分割合并成一行的对应

    #source = re.sub(r'\t', '    ', source) # 替换掉 tab 键为空格
    source = beautiful_jade(source)

    source = re.sub(r'((?:^|\n) *)else if ', '\g<1>elif ', source) # 替换else if->elif

    for func_name in ONE_LINE_FUNCTIONS:
        # 对单行特定函数，先处理为template的语法
        source = re.sub(r'([\r\n] *|^ *)(%s\(.*?\))(?= *[\r\n]|$)'%func_name, '\g<1>{{\g<2>}}', source) # 对单行的load函数的处理，避免被当成一个TagName

    new_source = process(source, compiler=Compiler)

    # 对头部代码的补充处理
    head_codes_search = re.search(r'<head>.*?</head>', new_source, re.S)
    if head_codes_search and 'set_content_type(' not in new_source: # 如果有设定content_type的，就不额外处理了
        if not new_source.startswith('<!'):
            new_source = "<!DOCTYPE html>\n"+new_source
        head_codes = head_codes_search.group()
        #if not re.search(r'<link.*?rel=["\']alternate["\']', head_codes, re.I): # 增加feed链接
        #    new_source = new_source.replace('<head>', '<head>%s' % FEED_HTML)
        if not re.search(r'<meta.*?http-equiv=[\'"]content-type[\'"]', head_codes, re.I): # 增加content_type声明
            new_source = new_source.replace('<head>', '<head>%s' % CONTENT_TYPE_HTML)

    return new_source



def jade_to_template(source, env=None):
    compiled_source = exchange(source)
    if env: # 指定了某个 env 的
        template = env.from_string(compiled_source)
    else:
        template = Template(compiled_source)
    template.source = compiled_source
    return template



def compile_jade_file(in_path):
    if not os.path.isfile(in_path):
        return
    if not in_path.endswith('.jade'):
        return
    out_path = in_path.replace('.jade', '.html')
    with open(in_path) as f:
        source_code = f.read()
    compiled_code = exchange(source_code)
    if isinstance(compiled_code, unicode):
        compiled_code = compiled_code.encode('utf8')
    with open(out_path, 'w') as f:
        f.write(compiled_code)



# coding: utf8
from __future__ import absolute_import
from flask import g

from farbox_lite.render.apis.data import get_doc, get_data, get_paginator as _get_paginator, get_connected_one
from farbox_lite.template_system.functions.namespace.utils.category import Category
from farbox_lite.utils import smart_unicode
from farbox_lite.utils.functional import cached_property
from farbox_lite.utils.web.cache_for_function import cache_result
from farbox_lite.utils.web.path import get_request_offset_path_without_prefix, get_request_path_without_prefix



class DocList(object):
    def __init__(self, namespace=''):
        self.pager_name = ''
        self.namespace = namespace
        self.min_per_page = 0

        self.sort = None # 可能设置某个排序的状态

        self.request_path = get_request_path_without_prefix()


    def set_min_per_page(self, min_per_page):
        if isinstance(min_per_page, int) and min_per_page >=0:
            self.min_per_page = min_per_page
        return '' # return nothing


    def set_sort(self, sort):
        if not isinstance(sort, (str, unicode, int, bool)):
            return # ignore
        if isinstance(sort, (str, unicode)):
            sort = smart_unicode(sort)[:10]
        self.sort = sort


    def get_paginator(self, name):
        return _get_paginator(name)

    def get_pager(self, name):
        return _get_paginator(name)


    @cached_property
    def all_categories(self):
        # 这个返回的是原生的数据对象，self.get_categories 则是 wrap 过的 Category 类实例化对象
        # all categories
        return get_data(type='folder', limit=1000, with_page=False) or []
        #return self.get_categories(sort='')
        # , level=

    @cache_result
    def get_categories(self, path='', level=0, sort='position', ignore_hidden=True):
        # 获取分类的通用函数
        kwargs = {}
        _categories = get_data(type='folder', limit=300, with_page=False, path=path, level=level, sort=sort, **kwargs) or []
        categories = []
        for c in _categories:
            if ignore_hidden: # path 以 _开头的，ignore
                if c.get('path', '').startswith('_'):
                    continue
            categories.append(Category(c))
        return categories


    @cached_property
    def category(self):
        # /xxx/<category_path>
        # 根据路径，获得当前的 category 对象
        if self.request_path == '/':
            return None
        doc = get_doc(path=get_request_offset_path_without_prefix(offset=1), match=True)
        if doc and doc.get('type') != 'folder':
            return None
        else:
            return Category(doc)
        
    @staticmethod
    def to_category(category_doc):
        if isinstance(category_doc, Category):
            return category_doc
        if category_doc and isinstance(category_doc, dict) and category_doc.get('type')=='folder':
            return Category(category_doc)
        return None


    @cached_property
    def categories(self):
        # 不是纯dict数据，而是Category包裹的
        category_objs = get_data(type='folder', limit=300, with_page=False, sort='position')
        categories = []
        for category_obj in category_objs:
            if category_obj.get('path', '').startswith('_'):
                continue
            categories.append(Category(category_obj))
        return categories

    @cached_property
    def sub_categories(self):
        parent_category = self.category
        if parent_category: # 当前有 category 的 url 命中，相当于子目录
            category_objs = get_data(types='folder', path=parent_category['path'], level=1, limit=300, with_page=False)
        else:
            category_objs = []
        categories = []
        for category_obj in category_objs:
            if category_obj.get('path', '').startswith('_'):
                continue
            categories.append(Category(category_obj))
        return categories



    @cached_property
    def context_doc(self):
        # 上下文对象
        doc_in_g = getattr(g, 'doc', None)
        return doc_in_g


    @cached_property
    def next_one(self):
        return self.get_pre_next_one(is_next=True)

    @cached_property
    def previous_one(self):
        return self.get_pre_next_one(is_next=False)

    @cached_property
    def pre_one(self):
        return self.previous_one

    @cached_property
    def current_folder_next_one(self):
        # 当前目录下的
        return self.get_pre_next_one(is_next=True, same_folder=True)

    @cached_property
    def current_folder_previous_one(self):
        # 当前目录下的
        return self.get_pre_next_one(is_next=False, same_folder=True)

    @cached_property
    def current_folder_pre_one(self):
        return self.current_folder_previous_one()

    def get_pre_next_one(self, is_next=True, same_folder=None):
        doc = self.context_doc
        if not doc:
            return # ignore

        date = doc.get('date')
        if not date:
            return

        if same_folder is None and doc.get('type') != 'post': # same_folder未指定，并且是post类型的，则全站
            same_folder = True

        related_doc = get_connected_one(doc=doc, same_folder=same_folder, position='>' if is_next else '<')
        return related_doc







#coding: utf8
from __future__ import absolute_import
import os
import datetime
from farbox_lite.compiler.utils import get_file_list
from farbox_lite.utils import get_relative_path
from farbox_lite.utils.mime import guess_type
import ujson as json

from .utils import is_real, is_a_hidden_path, get_sort_and_name_from_filepath

from .category import get_category_doc

# 此处不处理缓存，在 API 的逻辑中, 有 update 的逻辑, 因为有 cache_uuid 的逻辑


def filepath_to_file_doc(filepath, root=None):
    if not os.path.exists(filepath):
        return {}
    if not root:
        root = os.path.dirname(filepath)

    if os.path.isdir(filepath):
        # folder 直接进行添加处理的逻辑
        category_doc = get_category_doc(filepath, root=root)
        return category_doc

    relative_path = get_relative_path(filepath, root=root).strip().lower()
    slash_number = relative_path.count('/')
    mime_type = guess_type(filepath, 'application/octet-stream')
    file_type = 'file'
    if mime_type.startswith('image/'):
        file_type = 'image'

    sort, name = get_sort_and_name_from_filepath(filepath)
    just_name = os.path.splitext(name)[0]
    title = name
    if file_type == 'image':
        title = just_name

    date = datetime.datetime.fromtimestamp(os.path.getmtime(filepath))


    data = {}
    if relative_path.endswith('.json'):
        with open(filepath, 'rb') as f:
            raw_content = f.read()
        try:
            data = json.loads(raw_content)
        except:
            pass

    doc = dict(
        title = title,
        path = relative_path,
        filepath = filepath, # 记录完整的路径，后续 send_file 用的
        type = file_type,
        date = date,
        sort = sort,
        slash_number = slash_number,
        position = sort,

        data = data,
        id = filepath,
    )

    return doc


# todo 要不要把 category 也处理了, 放在 list 中?


def should_treat_as_a_file(filepath):
    l_filepath = filepath.lower()
    if '/template/' in l_filepath:
        return False
    if not os.path.isfile(filepath):
        return False
    if not is_real(filepath):
        return False
    if is_a_hidden_path(filepath):
        return False
    # at last
    return True



def get_compiled_file_docs(root):
    file_docs = []
    filepaths = get_file_list(root, split=False, filter_func=should_treat_as_a_file)
    for filepath in filepaths:
        doc = filepath_to_file_doc(filepath)
        if not doc:
            continue
        else:
            file_docs.append(doc)
    return file_docs








# coding: utf8
from __future__ import absolute_import
from flask import request, g

from farbox_lite.utils.web.cache_for_function import cache_result

from farbox_lite.utils import get_md5, smart_unicode, get_random_html_dom_id
from farbox_lite.utils.layout.utils import get_grid_factor
from farbox_lite.utils.functional import curry

from farbox_lite.utils.web.response import force_response
from farbox_lite.utils.web.request import get_user_agent
from farbox_lite.template_system.functions.namespace.html import Html


import re





@cache_result
def browser():
    # 对浏览器的判断，符合条件的才显示对应的内容
    def _func(family, **kwargs):
        caller = kwargs.pop('caller', None)
        if not caller or not hasattr(caller, '__call__'):
            return ''
        user_agent = get_user_agent()
        if not user_agent:
            return ''

        family = smart_unicode(family)
        totally_matched = False
        if family.startswith('@'):
            totally_matched = True
            family = family[1:]
        if totally_matched:
            return_html = family==user_agent.browser.family.lower()
        else:
            # 如果family 传入为空，则以下条件始终为 True
            return_html = family.lower() in user_agent.browser.family.lower()

        version = kwargs.get('version', None)
        if version and return_html: # 指定了 version 的
            browser_version = user_agent.browser.version_string
            if version == browser_version:
                pass
            elif version.startswith('>=') and version[2:]>=browser_version:
                pass
            elif version.startswith('>') and version[1:]>browser_version:
                pass
            elif version.startswith('<=') and version[2:]<=browser_version:
                pass
            elif version.startswith('<') and version[1:]<browser_version:
                pass
            else:
                return_html = False
        if return_html:
            html = caller()
            return html
        else:
            return '' # at last
    return _func


@cache_result
def create_dom():
    # 由于 jade 语法的显示，不能产生动态类型的 dom，所以增加了这个函数进行补足
    def _func(tag='div', *args, **kwargs):
        # *args 实际上是忽略的，没有任何作用
        caller = kwargs.pop('caller', None)
        inner_html = kwargs.pop('content', None) # 内部的 html 可以由 content 这个参数传入，或者其为空的时候，调用 caller
        extra_kwargs = kwargs.pop('extra', None)
        if extra_kwargs and isinstance(extra_kwargs, dict):
            kwargs.update(extra_kwargs)
        if inner_html:
            inner_html = smart_unicode(inner_html)
        if (not caller or not hasattr(caller, '__call__')) and not inner_html:
            return ''
        if not isinstance(tag, (str, unicode)):
            return ''
        if len(tag) > 20:
            return ''
        href = kwargs.pop('href', None)
        if tag == 'a' or href: # 有 href，也视为 a 元素
            tag = 'a'
            group_id = kwargs.pop('group_id', None)
        else:
            group_id = None
        properties = ''
        if tag == 'a' and href:
            create_a_dom = True
        else:
            create_a_dom = False
        for key, value in kwargs.items():
            if key in ['title', 'content'] or '_' in key:
                continue
            elif create_a_dom and key in ['href', 'class']:
                continue
            else:
                properties += ' %s="%s" '%(key, value)
        if inner_html is None and caller:
            inner_html = caller()
        elif not inner_html:
            inner_html = ''
        if create_a_dom: # 调用 auto_a, 可以实现一些基本预览的处理
            html = Html.auto_a(inner_html, href, a_title=kwargs.get('title'), class_name=kwargs.get('class'), group_id=group_id)
            if properties:
                html = html.replace('<a ', '<a %s '% properties, 1)
            return html
        html = '<%s %s>%s</%s>'  % (tag, properties, inner_html, tag)
        return html
    return _func



@cache_result
def footer(): # 兼容 FarBox、Bitcron
    def _func(*args, **kwargs):
        caller = kwargs.pop('caller', None)
        if not caller or not hasattr(caller, '__call__'):
            return ''
        inner_html = caller()
        g.footer = inner_html
        return inner_html
    return _func


# 兼容
@cache_result
def font():
    # +font(path, **name, title_path, title_name,)
    def _func(font_filepath, *args, **kwargs):
        caller = kwargs.pop('caller', None)
        if not caller or not hasattr(caller, '__call__') or not g.site:
            return ''
        embed_html = caller()
        return embed_html
    return _func


@cache_result
def pure(): # 对 pure css 的处理
    def _func(*args, **kwargs):
        do_grid = True
        caller = kwargs.pop('caller', None)
        if not caller or not hasattr(caller, '__call__'):
            return ''
        inner_html = caller()

        factors_devices = ['pure-u', # 总是,  # 全填满的时候相当于phone
                           'pure-u-sm', # ≥ 568px  # phone 横
                           'pure-u-md', # ≥ 768px  # ipad
                           'pure-u-lg', # ≥ 1024px # ipad 横 & 普通桌面
                           'pure-u-xl', # ≥ 1280px # 大桌面
                           ]
        html_class = smart_unicode(kwargs.get('class') or '')
        html_class_list = html_class.split('.')
        html_class = ' '.join(html_class_list)
        if len(args):
            for i, raw_factor in enumerate(args):
                grid_factor = get_grid_factor(raw_factor)
                grid_value, grid_base = grid_factor
                try: prefix = factors_devices[i]
                except: break
                grid_class = '%s-%s-%s' % (prefix, grid_value, grid_base)
                html_class += ' %s'%grid_class
            do_grid = False
        if do_grid: # 外部的 grid 包裹
            return '%s\n<div class="pure-g">\n%s\n</div>' % (Html.load('pure'), inner_html)
        else: # 某个具体 grid
            html_class = html_class.replace('"', '') + ' pure_cell'
            return '\n<div class="%s">\n%s\n</div>' % (html_class, inner_html)

    return _func





####### refer starts

def re_sub_refer(re_obj, handler=None):
    prefix = re_obj.group(1)
    url = re_obj.group(2)
    suffix = re_obj.group(3)
    new_line = False
    if prefix and suffix:
        if prefix.startswith('<span ') and 'md_line' in prefix and suffix == '</span>':
            new_line = True
        elif re.match(r'<div[ >]', prefix) and suffix == '</div>':
            new_line = True
        elif re.match(r'<p[ >]', prefix) and suffix=='</p>':
            new_line = True
    original_html = re_obj.group(0)
    new_html = original_html

    # 三个变量，逐个试过去...
    try:
        new_html = handler(url, original_html, new_line)
    except TypeError:
        try: new_html = handler(url, original_html)
        except TypeError:
            try: new_html = handler(url)
            except: pass
    except:
        pass
    return new_html

@cache_result
def refer():
    def _func(handler=None, refer_type='link', **kwargs):
        caller = kwargs.pop('caller', None)
        if not caller or not hasattr(caller, '__call__'):
            return ''
        inner_html = caller()
        if not handler or not hasattr(handler, '__call__'):
            return inner_html
        if refer_type in ['image', 'images']:
            # 图片
            inner_html = re.sub(r"""(<[^<>]+>)?<img [^<>]*?src=['"](.*?)['"][^<>]*?/?>(</\w+>)?""", curry(re_sub_refer, handler=handler), inner_html)
        else:
            # 链接
            inner_html = re.sub(r"""(<[^<>]+>)?<a [^<>]*?href=['"](.*?)['"][^<>]*?>.*?</a>(</\w+>)?""", curry(re_sub_refer, handler=handler), inner_html)
        return inner_html
    return _func


####### refer ends



#coding: utf8
from __future__ import absolute_import

# patch
from farbox_lite.template_system.monkey.patch import *

from flask import Flask
from farbox_lite.template_system.app_functions.after_request.cookie_handler import default_response_cookie_handler
from farbox_lite.template_system.app_functions.ba_request.time_cost import time_cost_handler
from farbox_lite.template_system.app_functions.before_request.basic import set_basic_vars
from farbox_lite.template_system.app_functions.utils import apply_after_request, apply_before_request

from farbox_lite.utils.error import capture_python_errors_to_local_error_log, capture_error
capture_python_errors_to_local_error_log()

from farbox_lite.render.route_view import web_route_view



app = Flask(__name__)


apply_before_request(app, time_cost_handler)
apply_after_request(app, time_cost_handler)

apply_after_request(app, default_response_cookie_handler)


apply_before_request(app, set_basic_vars)


@app.errorhandler(422)
def replace_response(error):
    # 强制 response 的替换，相当于拦截
    new_response = error.description
    return new_response




#coding: utf8
from __future__ import absolute_import
from jinja2.runtime import Context, Undefined, Macro
from jinja2.compiler import CodeGenerator
from werkzeug.exceptions import HTTPException

from farbox_lite.utils import smart_unicode
from farbox_lite.template_system.functions.namespace import default_variables


old_context_call = Context.call


# context resolve是全局性的hack
def context_resolve(self, key):
    if key in self.vars:
        return self.vars[key]
    if key in self.parent:
        return self.parent[key]

    key_lower = key.lower()
    if key in default_variables: # 全局的变量匹配，实际上都是函数
        return default_variables[key]
    elif key_lower in default_variables:
        return default_variables[key_lower]

    # at last
    return self.environment.undefined(name=key)



def context_call(self, context_obj, *args, **kwargs):
    # 一般只是本地的 jade 才会调用到，外部的实际走evn_call
    if not hasattr(context_obj, '__call__'): # 不可 call, 可能仅仅是 property
        return context_obj
    try:
        value = old_context_call(self, context_obj, *args, **kwargs)
        if value is None:
            value = Undefined()
        return value
    except HTTPException, e:
        raise  e
    except TypeError, e:
        if hasattr(e, 'message') and 'not callable' in e.message:
            # 被调用的函数不可被call，返回原始值
            return context_obj
        else:
            # sentry_client.captureException()
            message = getattr(e, 'message', None) or ''
            if message:
                return message
            error_info = 'context_call error:* %s, ** %s; %s' % (smart_unicode(args), smart_unicode(kwargs), message)
            return error_info
    except Exception, e:
        if isinstance(context_obj, Macro):
            pass
        else:
            # todo should captureException?
            pass

        message = getattr(e, 'message', None) or ''
        object_name = getattr(context_obj, '__name__', None) or ''
        if message:
            error_info = message
        else:
            error_info = 'context_call error: %s, * %s, ** %s; %s' % (object_name, smart_unicode(args), smart_unicode(kwargs), message)

        return '<error>%s</error>' % error_info




Context.resolve = context_resolve
Context.call = context_call


old_visit_name_func = CodeGenerator.visit_Name

def visit_name(self, node, frame):
    if node.name in default_variables and node.ctx=='load':
        # 默认变量最终转化为 call 的模式，因为本身就是函数，这样不会改变模板内代码整体的执行次序
        self.write("environment.call(context, l_%s)" % node.name)
    else:
        old_visit_name_func(self, node, frame)


CodeGenerator.visit_Name = visit_name




#coding: utf8
from __future__ import absolute_import
import os
from farbox_lite.compiler.posts.post import filepath_to_post_doc
from farbox_lite.compiler.utils import get_index_doc_filepath
from farbox_lite.compiler.templates import get_template_resources
from farbox_lite.utils import same_slash


site_docs_cache = {}


def get_site_doc_by_filepath_directly(site_path):
    # 没有 index.md .etc 时, 直接构建的
    if not os.path.isdir(site_path):
        return {}
    folder_name = os.path.split(site_path)[-1]
    site_doc = dict(
        title = folder_name,
        domains = [folder_name.strip().lower()],
        filepath = site_path
    )
    return site_doc



def raw_get_site_doc(site_path):
    site_path = same_slash(site_path)
    if not os.path.isdir(site_path):
        return {}
    index_doc_filepath = get_index_doc_filepath(site_path)
    if not index_doc_filepath:
        site_doc = get_site_doc_by_filepath_directly(site_path)
    else:
        site_doc = filepath_to_post_doc(index_doc_filepath)
    site_doc['type'] = 'site'
    site_doc['filepath'] = site_path
    site_doc['configs'] = site_doc.get('metadata') or {}
    site_doc['template_resources'] = get_template_resources(site_path)
    return site_doc



def get_site_doc(site_path):
    site_path = same_slash(site_path)
    if site_path in site_docs_cache:
        return site_docs_cache.get(site_path)
    else:
        site_doc = raw_get_site_doc(site_path)
        site_docs_cache[site_path] = site_doc
        return site_doc


def update_site_doc(site_path):
    site_doc = raw_get_site_doc(site_path)
    site_docs_cache[site_path] = site_doc





#coding: utf8
from __future__ import absolute_import
import io
import os
import re
from flask import abort, request, send_file
from flask.ctx import HTTPException
from jinja2.exceptions import TemplateNotFound

from farbox_lite.template_system.env import farbox_lite_env
from farbox_lite.render.apis.site import get_site_from_request, get_sites
from farbox_lite.render.apis.all_docs import get_doc_by_path
from farbox_lite.utils import smart_str, smart_unicode
from farbox_lite.utils.mime import guess_type
from farbox_lite.utils.web.path import get_path
from farbox_lite.utils.error import capture_error




def render_template_static_file():
    path = request.path.strip('/')
    if not path.startswith('template/'):
        return # ignore
    if path.endswith('.jade'):
        return
    site = get_site_from_request()
    if not site:
        return
    site_root = site['filepath']
    prefix, ext = os.path.splitext(path)
    if ext in ['.scss', '.less']:
        path = prefix + '.css'
    elif ext == '.coffee':
        path = prefix + '.js'
    mime_type = guess_type(path) or 'application/octet-stream'
    template_resources = site.get('template_resources')
    if not template_resources or not isinstance(template_resources, dict):
        return
    template_path = path.replace('template/', '', 1).strip('/').lower()
    abs_filepath = os.path.join(site_root, path)
    if template_path not in template_resources:
        return
    if os.path.isfile(abs_filepath):
        return send_file(abs_filepath, mimetype=mime_type)
    else:
        raw_content = smart_str(template_resources.get(template_path))
        file_response = send_file(io.BytesIO(raw_content), mimetype=mime_type)
        return file_response



def render_common_static_file(html_after_render_func=None):
    path = request.path.strip('/').lower()
    raw_path = request.path.strip('/')
    if request.path.endswith('/'):
        raw_path = request.path + 'index.html'
    raw_path = raw_path.strip('/')
    if not raw_path:
        raw_path = 'index.html'
    site = get_site_from_request()
    if not site:
        return
    site_root = site['filepath']
    abs_filepath = os.path.join(site_root, raw_path)
    if os.path.isfile(abs_filepath):
        mime_type = guess_type(raw_path) or 'application/octet-stream'
        if html_after_render_func and re.search(r'\.html?$', raw_path, flags=re.I):
            with open(abs_filepath, 'rb') as f:
                raw_html_content = f.read()
            raw_html_content = smart_unicode(raw_html_content)
            html_content = html_after_render_func(raw_html_content)
            return send_file(io.BytesIO(smart_str(html_content)), mimetype=mime_type)
        else:
            return send_file(abs_filepath, mimetype=mime_type)
    else:
        file_doc = get_doc_by_path(path, root=site_root)
        if file_doc:
            mime_type = guess_type(path) or 'application/octet-stream'
            filepath = file_doc.get('filepath')
            file_type = file_doc.get('type')
            if file_type in ['file', 'image'] and filepath and os.path.isfile(filepath):
                return send_file(filepath, mimetype=mime_type)



def web_route_view(after_render_func=None, static_html_after_render_func=None):
    # step1, 先处理 template 下的静态资源
    template_static_file_response = render_template_static_file()
    if template_static_file_response:
        return template_static_file_response

    request_path = get_path() # also set g.page
    try:
        template = farbox_lite_env.get_template(request_path)
        if template:
            html = template.render()
            if after_render_func and hasattr(after_render_func, '__call__'):
                html = after_render_func(html)
            return html
        else:
            abort(404)
    except TemplateNotFound:
        static_file_response = render_common_static_file(html_after_render_func=static_html_after_render_func)
        if static_file_response:
            return static_file_response
        else:
            sites = get_sites()
            sites_count = len(sites) if sites else 0
            abort(404, 'TemplateNotFound  or no static file matched, sites: %s' % sites_count)
    except HTTPException, e:
        raise e
    except:
        capture_error()
        abort(500)


#coding: utf8
from __future__ import absolute_import
from farbox_lite.utils import to_unicode, get_md5
from jinja2.loaders import  BaseLoader
from jinja2.environment import Template
from jinja2.utils import internalcode
from jinja2.exceptions import TemplateNotFound
from jinja2 import sandbox
import os

from farbox_lite.compiler.templates import update_template_resources
from farbox_lite.template_system.utils import SafeUndefined, Undefined
from farbox_lite.render.apis.site import get_site_from_request
from farbox_lite.template_system.monkey.attr import env_call, env_getattr, env_handle_exception


def return_error_message(message):
    if 'error_ignored' in message:
        error_info = ''
    else:
        error_info = '<span style="color:red" class="template_api_error">&lt;%s&gt:</span>' % message
    return error_info




def get_template_source(template_name, hit_newest=True):
    site = get_site_from_request()
    if not site:
        return
    site_root = site['filepath']
    template_resources = site.get('template_resources')
    if not template_resources or not isinstance(template_resources, dict):
        return

    template_resource_mtimes = template_resources.get('_')
    if not isinstance(template_resource_mtimes, dict):
        template_resource_mtimes = {}

    template_name = template_name.strip('"\'').strip('/').lower()
    if not template_name: # 首页
        template_name = 'index.html'
    if template_name.endswith('.jade'):
        template_name = template_name.replace('.jade', '.html')
    if not template_name.endswith('.html'): # 比如 import a.b
        template_name = template_name.replace('.', '/')
        template_name += '.html'

    source = ''
    template_name_to_hit = None
    template_name_parts = os.path.splitext(template_name)[0].split('/')
    for length in range(len(template_name_parts), 0, -1):
        template_name_to_hit = '/'.join(template_name_parts[:length]) + '.html'
        if template_name_to_hit in template_resources:
            source = to_unicode(template_resources.get(template_name_to_hit))
            break

    if hit_newest and source and template_name_to_hit:
        template_resource_mtime_info = template_resource_mtimes.get(template_name_to_hit)
        if template_resource_mtime_info:
            template_filepath, mtime = template_resource_mtime_info
            if os.path.isfile(template_filepath):
                new_mtime = os.path.getmtime(template_filepath)
                if new_mtime != mtime:
                    site['template_resources'] = update_template_resources(site_root)
                    return get_template_source(template_name, hit_newest=False)

    # at last
    return source


class FarboxLiteTemplateLoader(BaseLoader):
    def get_source(self, environment, template_name):
        template_source = get_template_source(template_name)
        if template_source:
            return template_source, template_name, lambda: True
        else:
            raise TemplateNotFound(template_name)

    @internalcode
    def load(self, environment, name, globals=None):
        globals = globals or {}
        source, filename, uptodate = self.get_source(environment, name)
        code = environment.compile(source, name, filename)
        template = environment.template_class.from_code(environment, code, globals, uptodate)
        template.source = source
        return template




class FarboxLiteEnvironment(sandbox.SandboxedEnvironment): #  Environment
    intercepted_binops = frozenset(['**', '*', '/', '+'])
    def __init__(self, *args, **kwargs):
        kwargs['autoescape'] = False
        kwargs['auto_reload'] = False
        kwargs['cache_size'] = 50000
        kwargs['extensions'] = ['jinja2.ext.do']
        super(FarboxLiteEnvironment, self).__init__(*args, **kwargs)
        self.undefined = SafeUndefined
        self.loader =  FarboxLiteTemplateLoader()

    @staticmethod
    def template_not_found(name):
        content = return_error_message("template error: can't find the template named `%s`" % name)
        return Template(content)

    def get_template(self, name, parent=None, globals=None):
        if isinstance(name, Template):
            return name
        globals = self.make_globals(globals)

        is_parent = bool(not parent)

        template_source = get_template_source(name)
        if template_source:
            template_md5_key = get_md5(template_source)
            template = self.cache.get(template_md5_key)
            if template is None:
                template = self.loader.load(environment=self, name=name, globals=globals)
                self.cache[template_md5_key] = template
            return template
        else:
            if is_parent:
                raise TemplateNotFound(name)
            else:
                return self.template_not_found(name)


    def getattr(self, obj, attribute):
        if isinstance(obj, Undefined):
            return SafeUndefined()
        return env_getattr(self, obj, attribute, sandbox=True)


    def call(self, context, context_obj, *args, **kwargs):
        """Call an object from sandboxed code."""
        # the double prefixes are to avoid double keyword argument
        # errors when proxying the call.
        return env_call(self, context, context_obj, *args, **kwargs)


    def handle_exception(self, exc_info=None, rendered=False, source_hint=None):
        return env_handle_exception(self, exc_info, rendered, source_hint)


farbox_lite_env = FarboxLiteEnvironment()



# coding: utf8
from __future__ import absolute_import
from flask import g
import re, os
from farbox_lite.utils.functional import  curry
from farbox_lite.utils.mime import guess_type


def get_url_with_prefix(url, force=False, prefix=None):
    # 转换 url (url_path), 自动补全前缀
    # force 的话，就会强制处理, 即使不以 / 开头的，都会尝试转换
    if not isinstance(url, (str, unicode)):
        return url
    if not url.startswith('/') and not force:
        # 仅仅处理当前网站的URL，以避免混淆
        return url
    if prefix is None:
        prefix = getattr(g, 'prefix', '')
    if prefix and isinstance(prefix, (str,unicode)):
        raw_url = url
        prefix = prefix.strip().strip('/')
        url = url.lstrip('/')
        if url.lower().startswith(prefix+'/') or url.lower() == prefix:
            # 本身已经是前缀补足了
            # 所以，有一个问题，在 /prefix 下，不能存在 一个/prefix/prefix 的意图，这不会转换
            return raw_url
        if url:
            new_url = '/%s/%s' % (prefix, url)
        else: # 相当原始的首页 /
            new_url = '/%s' % prefix
        return new_url
    else:
        return url


def get_url_without_prefix(url, prefix=None):
    # url 去除 prefix 后， 一般以 '/' 开头
    if not isinstance(url, (str, unicode)):
        return url # ignore
    url_startswith_dash = url.startswith('/')
    raw_url = url
    if prefix is None:
        prefix = getattr(g, 'prefix', '')
    if prefix and isinstance(prefix, (str,unicode)):
        url = url.lstrip('/')
        prefix = prefix.strip().strip('/')
        if url.startswith(prefix+'/') or url == prefix:
            first_char = '/' if url_startswith_dash else ''
            url =  first_char + url.replace(prefix, '', 1).lstrip('/')
            # 确保模拟 request.path 类似的逻辑，一般可以保证 / 开头的
            return url
    return raw_url



# 只处理 / 开头的url, 视为 site 内部的 url
site_a_dom_re_compiler = re.compile(r'(<a[^<]+href=[\'"])(/.*?)([\'"][^<]*>.*?</a>)', flags=re.S|re.I)


def auto_prefix_a_re_obj_replacer(re_obj, prefix_url, ignore_file_url=True, ignore_keywords=('nav__item',)):
    part1 = re_obj.group(1)
    current_url = re_obj.group(2)
    part3 = re_obj.group(3)
    raw_a_html = re_obj.group()
    if '#' in current_url: # 有 # 存在的，则不进行转义、自动补全
        return raw_a_html
    for ignore_keyword in ignore_keywords:
        # 有些关键词，会自动不进行补全，比如作为自动生成的 nav
        if ignore_keyword in raw_a_html:
            return raw_a_html
    if ignore_file_url and current_url:
        # 直接的文件输出地址，不做补全，避免画蛇添足, 但也只是大概的估算
        file_type = guess_type(current_url)
        # link_in_site 是非常特殊的字符，如果存在，则认为这个也是要处理
        if 'link_in_site' not in raw_a_html:
            if file_type.startswith('image/') or file_type.startswith('text/') or file_type.startswith('application/'):
                return raw_a_html
    new_url = get_url_with_prefix(current_url, prefix=prefix_url)
    return '%s%s%s' % (part1, new_url, part3)



def auto_complete_prefix_url_for_html(html, prefix_url=None, ignore_file_url=True):
    if not isinstance(prefix_url, (str, unicode)):
        return html
    prefix_url = prefix_url.strip().strip('/')
    if not prefix_url:
        return html
    new_html = site_a_dom_re_compiler.sub(curry(auto_prefix_a_re_obj_replacer, prefix_url=prefix_url, ignore_file_url=ignore_file_url), html)
    return new_html
